"""Interface for default PDF sets"""

import logging
from typing import Any, Callable, Dict, List, Optional, Text, Tuple, Union

from autograd import value_and_grad, hessian, jacobian
from autograd import numpy as np
from scipy.optimize import NonlinearConstraint

from spey._version import __version__
from spey.backends.distributions import ConstraintModel, MainModel
from spey.base import BackendBase, ModelConfig
from spey.helper_functions import covariance_to_correlation
from spey.optimizer import fit
from spey.utils import ExpectationType

from .third_moment import third_moment_expansion
from .uncertainty_synthesizer import signal_uncertainty_synthesizer

# pylint: disable=E1101,E1120
log = logging.getLogger("Spey")

# pylint: disable=W1203


class DefaultPDFBase(BackendBase):
    """
    Default PDF backend base

    Args:
        signal_yields (``np.ndarray``): signal yields
        background_yields (``np.ndarray``): background yields
        data (``np.ndarray``): observed yields
        covariance_matrix (``np.ndarray``): covariance matrix. The dimensionality of each axis has
          to match with ``background_yields``, ``signal_yields``, and ``data`` inputs.

          .. warning::

            The diagonal terms of the covariance matrix involves squared absolute background
            uncertainties. In case of uncorralated bins user should provide a diagonal matrix
            with squared background uncertainties.

        signal_uncertainty_configuration (``Dict[Text, Any]]``, default ``None``): Configuration
          input for signal uncertainties

          * absolute_uncertainties (``List[float]``): Absolute uncertainties for the signal
          * absolute_uncertainty_envelops (``List[Tuple[float, float]]``): upper and lower
              uncertainty envelops
          * correlation_matrix (``List[List[float]]``): Correlation matrix
          * third_moments (``List[float]``): diagonal elemetns of the third moment

    .. note::

        To enable a differentiable statistical model, all inputs are wrapped with
        :func:`autograd.numpy.array` function.
    """

    name: Text = "default_pdf.base"
    """Name of the backend"""
    version: Text = __version__
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = __version__
    """Spey version required for the backend"""

    __slots__ = [
        "_model",
        "_main_model",
        "_constraint_model",
        "constraints",
        "signal_uncertainty_configuration",
    ]

    def __init__(
        self,
        signal_yields: np.ndarray,
        background_yields: np.ndarray,
        data: np.ndarray,
        covariance_matrix: Optional[
            Union[np.ndarray, Callable[[np.ndarray], np.ndarray]]
        ] = None,
        signal_uncertainty_configuration: Optional[Dict[Text, Any]] = None,
    ):
        self.data = np.array(data, dtype=np.float64)
        self.signal_yields = np.array(signal_yields, dtype=np.float64)
        self.background_yields = np.array(background_yields, dtype=np.float64)
        self.covariance_matrix = (
            np.array(covariance_matrix, dtype=np.float64)
            if not callable(covariance_matrix) and covariance_matrix is not None
            else covariance_matrix
        )
        if signal_uncertainty_configuration is None:
            self.signal_uncertainty_configuration = {}
        else:
            self.signal_uncertainty_configuration = signal_uncertainty_synthesizer(
                signal_yields=self.signal_yields,
                **signal_uncertainty_configuration,
                domain=slice(len(background_yields) + 1, None),
            )

        minimum_poi = -np.inf
        if self.is_alive:
            minimum_poi = -np.min(
                self.background_yields[self.signal_yields > 0.0]
                / self.signal_yields[self.signal_yields > 0.0]
            )
        log.debug(f"Min POI set to : {minimum_poi}")

        self._main_model = None
        self._constraint_model = None
        self.constraints = []
        """Constraints to be used during optimisation process"""

        self._config = ModelConfig(
            poi_index=0,
            minimum_poi=minimum_poi,
            suggested_init=[1.0] * (len(data) + 1)
            + (signal_uncertainty_configuration is not None)
            * ([1.0] * len(signal_yields)),
            suggested_bounds=[(minimum_poi, 10)]
            + [(None, None)] * len(data)
            + (signal_uncertainty_configuration is not None)
            * ([(None, None)] * len(signal_yields)),
        )

    @property
    def is_alive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        return np.any(self.signal_yields > 0.0)

    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: float = 10.0
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """
        if allow_negative_signal and poi_upper_bound == 10.0:
            return self._config

        return ModelConfig(
            self._config.poi_index,
            self._config.minimum_poi,
            self._config.suggested_init,
            [(0, poi_upper_bound)] + self._config.suggested_bounds[1:],
        )

    @property
    def constraint_model(self) -> ConstraintModel:
        """retreive constraint model distribution"""
        if self._constraint_model is None:
            corr = covariance_to_correlation(self.covariance_matrix)
            self._constraint_model = ConstraintModel(
                [
                    {
                        "distribution_type": "multivariatenormal",
                        "args": [np.zeros(len(self.data)), corr],
                        "kwargs": {"domain": slice(1, None)},
                    }
                ]
                + self.signal_uncertainty_configuration.get("constraint", [])
            )
        return self._constraint_model

    @property
    def main_model(self) -> MainModel:
        """retreive the main model distribution"""
        if self._main_model is None:
            A = self.background_yields
            B = np.sqrt(np.diag(self.covariance_matrix))

            def lam(pars: np.ndarray) -> np.ndarray:
                """
                Compute lambda for Main model with third moment expansion.
                For details see above eq 2.6 in :xref:`1809.05548`

                Args:
                    pars (``np.ndarray``): nuisance parameters

                Returns:
                    ``np.ndarray``:
                    expectation value of the poisson distribution with respect to
                    nuisance parameters.
                """
                return pars[0] * self.signal_yields + A + B * pars[slice(1, len(B) + 1)]

            def constraint(pars: np.ndarray) -> np.ndarray:
                """Compute constraint term"""
                return A + B * pars[slice(1, len(B) + 1)]

            jac_constr = jacobian(constraint)

            self.constraints.append(
                NonlinearConstraint(constraint, 0.0, np.inf, jac=jac_constr)
            )

            if self.signal_uncertainty_configuration.get("lambda", None) is not None:
                signal_lambda = self.signal_uncertainty_configuration["lambda"]

                def poiss_lamb(pars: np.ndarray) -> np.ndarray:
                    """combined lambda expression"""
                    return lam(pars) + signal_lambda(pars)

            else:
                poiss_lamb = lam

            self._main_model = MainModel(poiss_lamb)

        return self._main_model

    def get_objective_function(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[np.ndarray] = None,
        do_grad: bool = True,
    ) -> Callable[[np.ndarray], Union[Tuple[float, np.ndarray], float]]:
        r"""
        Objective function i.e. twice negative log-likelihood, :math:`-2\log\mathcal{L}(\mu, \theta)`

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``np.ndarray``, default ``None``): input data that to fit
            do_grad (``bool``, default ``True``): If ``True`` return objective and its gradient
              as ``tuple`` if ``False`` only returns objective function.

        Returns:
            ``Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]``:
            Function which takes fit parameters (:math:`\mu` and :math:`\theta`) and returns either
            objective or objective and its gradient.
        """
        current_data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        )
        data = current_data if data is None else data
        log.debug(f"Data: {data}")

        def negative_loglikelihood(pars: np.ndarray) -> np.ndarray:
            """Compute twice negative log-likelihood"""
            return -self.main_model.log_prob(
                pars, data[: len(self.data)]
            ) - self.constraint_model.log_prob(pars)

        if do_grad:
            return value_and_grad(negative_loglikelihood, argnum=0)

        return negative_loglikelihood

    def get_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[np.array] = None,
    ) -> Callable[[np.ndarray, np.ndarray], float]:
        r"""
        Generate function to compute :math:`\log\mathcal{L}(\mu, \theta)` where :math:`\mu` is the
        parameter of interest and :math:`\theta` are nuisance parameters.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``np.array``, default ``None``): input data that to fit

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and computes
            :math:`\log\mathcal{L}(\mu, \theta)`.
        """
        current_data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        )
        data = current_data if data is None else data
        log.debug(f"Data: {data}")

        return lambda pars: self.main_model.log_prob(
            pars, data[: len(self.data)]
        ) + self.constraint_model.log_prob(pars)

    def get_hessian_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[np.ndarray] = None,
    ) -> Callable[[np.ndarray], float]:
        r"""
        Currently Hessian of :math:`\log\mathcal{L}(\mu, \theta)` is only used to compute
        variance on :math:`\mu`. This method returns a callable function which takes fit
        parameters (:math:`\mu` and :math:`\theta`) and returns Hessian.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``np.ndarray``, default ``None``): input data that to fit

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and
            returns Hessian of :math:`\log\mathcal{L}(\mu, \theta)`.
        """
        current_data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        )
        data = current_data if data is None else data
        log.debug(f"Data: {data}")

        def log_prob(pars: np.ndarray) -> np.ndarray:
            """Compute log-probability"""
            return self.main_model.log_prob(
                pars, data[: len(self.data)]
            ) + self.constraint_model.log_prob(pars)

        return hessian(log_prob, argnum=0)

    def get_sampler(self, pars: np.ndarray) -> Callable[[int], np.ndarray]:
        r"""
        Retreives the function to sample from.

        Args:
            pars (``np.ndarray``): fit parameters (:math:`\mu` and :math:`\theta`)
            include_auxiliary (``bool``): wether or not to include auxiliary data
              coming from the constraint model.

        Returns:
            ``Callable[[int, bool], np.ndarray]``:
            Function that takes ``number_of_samples`` as input and draws as many samples
            from the statistical model.
        """

        def sampler(sample_size: int, include_auxiliary: bool = True) -> np.ndarray:
            """
            Fucntion to generate samples.

            Args:
                sample_size (``int``): number of samples to be generated.
                include_auxiliary (``bool``): wether or not to include auxiliary data
                    coming from the constraint model.

            Returns:
                ``np.ndarray``:
                generated samples
            """
            sample = self.main_model.sample(pars, sample_size)

            if include_auxiliary:
                constraint_sample = self.constraint_model.sample(pars[1:], sample_size)
                sample = np.hstack([sample, constraint_sample])

            return sample

        return sampler

    def expected_data(
        self, pars: List[float], include_auxiliary: bool = True
    ) -> List[float]:
        r"""
        Compute the expected value of the statistical model

        Args:
            pars (``List[float]``): nuisance, :math:`\theta` and parameter of interest,
              :math:`\mu`.
            include_auxiliary (``bool``): wether or not to include auxiliary data
              coming from the constraint model.

        Returns:
            ``List[float]``:
            Expected data of the statistical model
        """
        data = self.main_model.expected_data(pars)

        if include_auxiliary:
            data = np.hstack([data, self.constraint_model.expected_data()])
        return data


class UncorrelatedBackground(DefaultPDFBase):
    r"""
    Interface for uncorrelated background uncertainties.
    This simple backend is designed to handle single or multi-bin statistical models
    with uncorrelated uncertainties. Inputs has to be given as list of ``NumPy`` array where
    each input should include same number of regions. It assumes absolute uncertainties
    on the background sample e.g. for a background sample yield reported as
    :math:`3.1\pm0.5` the background yield is ``3.1`` and the absolute uncertainty is
    ``0.5``. It forms a combination of normal and poisson distributions to from the input
    data where the log-probability is computed as sum of all normal and poisson distributions.

    Args:
        signal_yields (``List[float]``): signal yields
        background_yields (``List[float]``): background yields
        data (``List[int]``): observations
        absolute_uncertainties (``List[float]``): absolute uncertainties on the background
        signal_uncertainty_configuration (``Dict[Text, Any]]``, default ``None``): Configuration
          input for signal uncertainties

          * absolute_uncertainties (``List[float]``): Absolute uncertainties for the signal
          * absolute_uncertainty_envelops (``List[Tuple[float, float]]``): upper and lower
              uncertainty envelops
          * correlation_matrix (``List[List[float]]``): Correlation matrix
          * third_moments (``List[float]``): diagonal elemetns of the third moment

    .. note::

        Each input should have the same dimensionality, i.e. if ``data`` has three regions,
        ``signal_yields``, ``background_yields`` and ``absolute_uncertainties`` inputs should
        have three regions as well.

    Example:

    .. code:: python3

        >>> import spey
        >>> stat_wrapper = spey.get_backend('default_pdf.uncorrelated_background')

        >>> data = [1, 3]
        >>> signal = [0.5, 2.0]
        >>> background = [2.0, 2.8]
        >>> background_unc = [1.1, 0.8]

        >>> stat_model = stat_wrapper(
        ...     signal, background, data, background_unc, analysis="multi-bin", xsection=0.123
        ... )
        >>> print("1-CLs : %.3f" % tuple(stat_model.exclusion_confidence_level()))
    """

    name: Text = "default_pdf.uncorrelated_background"
    """Name of the backend"""
    version: Text = DefaultPDFBase.version
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = DefaultPDFBase.spey_requires
    """Spey version required for the backend"""

    def __init__(
        self,
        signal_yields: List[float],
        background_yields: List[float],
        data: List[int],
        absolute_uncertainties: List[float],
        signal_uncertainty_configuration: Optional[Dict[Text, Any]] = None,
    ):
        super().__init__(
            signal_yields=signal_yields,
            background_yields=background_yields,
            data=data,
            covariance_matrix=None,
            signal_uncertainty_configuration=signal_uncertainty_configuration,
        )

        B = np.array(absolute_uncertainties)

        self._constraint_model: ConstraintModel = ConstraintModel(
            [
                {
                    "distribution_type": "normal",
                    "args": [np.zeros(len(self.data)), np.ones(len(B))],
                    "kwargs": {"domain": slice(1, len(B) + 1)},
                }
            ]
            + self.signal_uncertainty_configuration.get("constraint", [])
        )

        def lam(pars: np.ndarray) -> np.ndarray:
            """Compute lambda for Main model"""
            return (
                self.background_yields
                + pars[slice(1, len(B) + 1)] * B
                + pars[0] * self.signal_yields
            )

        def constraint(pars: np.ndarray) -> np.ndarray:
            """Compute the constraint term"""
            return self.background_yields + pars[slice(1, len(B) + 1)] * B

        jac_constr = jacobian(constraint)

        self.constraints.append(
            NonlinearConstraint(constraint, 0.0, np.inf, jac=jac_constr)
        )

        if self.signal_uncertainty_configuration.get("lambda", None) is not None:
            signal_lambda = self.signal_uncertainty_configuration["lambda"]

            def poiss_lamb(pars: np.ndarray) -> np.ndarray:
                """combined lambda expression"""
                return lam(pars) + signal_lambda(pars)

        else:
            poiss_lamb = lam

        self._main_model = MainModel(poiss_lamb)


class CorrelatedBackground(DefaultPDFBase):
    r"""
    Correlated multi-region statistical model.
    The correlation between each nuisance parameter has been captured via
    Multivariate Normal distribution and the log-probability distribution is
    combination of Multivariate Normal along with Poisson distribution.
    The Multivariate Normal distribution is constructed by the help
    of a covariance matrix provided by the user which captures the
    uncertainties and background correlations between each histogram bin.
    The probability distribution of a simplified likelihood can be formed as follows;

    .. math::

        \mathcal{L}_{SL}(\mu,\theta) = \underbrace{\left[\prod_i^N {\rm Poiss}\left(n^i_{obs}
        | \lambda_i(\mu, \theta)\right) \right]}_{\rm main\ model}
        \cdot \underbrace{\mathcal{N}(\theta | 0, \rho)}_{\rm constraint\ model}

    Here the first term is the so-called main model based on Poisson distribution centred around
    :math:`\lambda_i(\mu, \theta) = \mu n^i_{sig} + \theta + n^i_{bkg}` and the second term is the
    multivariate normal distribution centred around zero with the correlation matrix
    :math:`\rho`.

    Args:
        signal_yields (``np.ndarray``): signal yields
        background_yields (``np.ndarray``): background yields
        data (``np.ndarray``): observations
        covariance_matrix (``np.ndarray``): covariance matrix (square matrix)
        signal_uncertainty_configuration (``Dict[Text, Any]]``, default ``None``): Configuration
          input for signal uncertainties

          * absolute_uncertainties (``List[float]``): Absolute uncertainties for the signal
          * absolute_uncertainty_envelops (``List[Tuple[float, float]]``): upper and lower
              uncertainty envelops
          * correlation_matrix (``List[List[float]]``): Correlation matrix
          * third_moments (``List[float]``): diagonal elemetns of the third moment

    .. note::

        Each input should have the same dimensionality, i.e. if ``data`` has three regions,
        ``signal_yields`` and ``background_yields`` inputs should have three regions as well.
        Additionally ``covariance_matrix`` is expected to be square matrix, thus for a three
        region statistical model it is expected to be 3x3 matrix.

    Example:

    .. code:: python3

        >>> import spey

        >>> stat_wrapper = spey.get_backend('default_pdf.correlated_background')

        >>> signal_yields = [12.0, 11.0]
        >>> background_yields = [50.0, 52.0]
        >>> data = [51, 48]
        >>> covariance_matrix = [[3.,0.5], [0.6,7.]]

        >>> statistical_model = stat_wrapper(signal_yields,background_yields,data,covariance_matrix)
        >>> print(statistical_model.exclusion_confidence_level())
    """

    name: Text = "default_pdf.correlated_background"
    """Name of the backend"""
    version: Text = DefaultPDFBase.version
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = DefaultPDFBase.spey_requires
    """Spey version required for the backend"""

    def __init__(
        self,
        signal_yields: np.ndarray,
        background_yields: np.ndarray,
        data: np.ndarray,
        covariance_matrix: np.ndarray,
        signal_uncertainty_configuration: Optional[Dict[Text, Any]] = None,
    ):
        super().__init__(
            signal_yields=signal_yields,
            background_yields=background_yields,
            data=data,
            covariance_matrix=covariance_matrix,
            signal_uncertainty_configuration=signal_uncertainty_configuration,
        )

        assert self.main_model is not None, "Unable to build the main model"
        assert self.constraint_model is not None, "Unable to build the constraint model"


class ThirdMomentExpansion(DefaultPDFBase):
    r"""
    Simplified likelihood interface with third moment expansion.
    Third moment expansion follows simplified likelihood construction
    and modifies the :math:`\lambda` and :math:`\Sigma`. Using the expected
    background yields, :math:`m^{(1)}_i`, diagonal elements of the third moments,
    :math:`m^{(3)}_i` and the covariance matrix, :math:`m^{(2)}_{ij}`, one
    can write a modified correlation matrix and :math:`\lambda` function as follows

    .. math::

        C_i &= -sign(m^{(3)}_i) \sqrt{2 m^{(2)}_{ii}} \cos\left( \frac{4\pi}{3} +
        \frac{1}{3}\arctan\left(\sqrt{ \frac{8(m^{(2)}_{ii})^3}{(m^{(3)}_i)^2} - 1}\right) \right)

        B_i &= \sqrt{m^{(2)}_{ii} - 2 C_i^2}

        A_i &=  m^{(1)}_i - C_i

        \rho_{ij} &= \frac{1}{4C_iC_j} \left( \sqrt{(B_iB_j)^2 + 8C_iC_jm^{(2)}_{ij}} - B_iB_j \right)

    which further modifies :math:`\lambda_i(\mu, \theta) = \mu n^i_{sig} + A_i + B_i \theta_i + C_i \theta_i^2`
    and the multivariate normal has been modified via the inverse of the correlation matrix,
    :math:`\mathcal{N}(\theta | 0, \rho^{-1})`. See :xref:`1809.05548` Sec. 2 for details.

    Args:
        signal_yields (``np.ndarray``): signal yields
        background_yields (``np.ndarray``): background yields
        data (``np.ndarray``): observations
        covariance_matrix (``np.ndarray``): covariance matrix (square matrix)
        third_moment (``np.ndarray``): third moment for each region.
        signal_uncertainty_configuration (``Dict[Text, Any]]``, default ``None``): Configuration
          input for signal uncertainties

          * absolute_uncertainties (``List[float]``): Absolute uncertainties for the signal
          * absolute_uncertainty_envelops (``List[Tuple[float, float]]``): upper and lower
              uncertainty envelops
          * correlation_matrix (``List[List[float]]``): Correlation matrix
          * third_moments (``List[float]``): diagonal elemetns of the third moment

    .. note::

        Each input should have the same dimensionality, i.e. if ``data`` has three regions,
        ``signal_yields`` and ``background_yields`` inputs should have three regions as well.
        Additionally ``covariance_matrix`` is expected to be square matrix, thus for a three
        region statistical model it is expected to be 3x3 matrix. Following these,
        ``third_moment`` should also have three inputs.
    """

    name: Text = "default_pdf.third_moment_expansion"
    """Name of the backend"""
    version: Text = DefaultPDFBase.version
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = DefaultPDFBase.spey_requires
    """Spey version required for the backend"""
    doi: List[Text] = ["10.1007/JHEP04(2019)064"]
    """Citable DOI for the backend"""
    arXiv: List[Text] = ["1809.05548"]
    """arXiv reference for the backend"""

    def __init__(
        self,
        signal_yields: np.ndarray,
        background_yields: np.ndarray,
        data: np.ndarray,
        covariance_matrix: np.ndarray,
        third_moment: np.ndarray,
        signal_uncertainty_configuration: Optional[Dict[Text, Any]] = None,
    ):
        third_moments = np.array(third_moment)

        super().__init__(
            signal_yields=signal_yields,
            background_yields=background_yields,
            data=data,
            covariance_matrix=covariance_matrix,
            signal_uncertainty_configuration=signal_uncertainty_configuration,
        )

        A, B, C, corr = third_moment_expansion(
            self.background_yields, self.covariance_matrix, third_moments, True
        )

        def lam(pars: np.ndarray) -> np.ndarray:
            """
            Compute lambda for Main model with third moment expansion.
            For details see above eq 2.6 in :xref:`1809.05548`

            Args:
                pars (``np.ndarray``): nuisance parameters

            Returns:
                ``np.ndarray``:
                expectation value of the poisson distribution with respect to
                nuisance parameters.
            """
            nI = (
                A
                + B * pars[slice(1, len(B) + 1)]
                + C * np.square(pars[slice(1, len(B) + 1)])
            )
            return pars[0] * self.signal_yields + nI

        def constraint(pars: np.ndarray) -> np.ndarray:
            """Compute constraint term"""
            return (
                A
                + B * pars[slice(1, len(B) + 1)]
                + C * np.square(pars[slice(1, len(B) + 1)])
            )

        jac_constr = jacobian(constraint)

        self.constraints.append(
            NonlinearConstraint(constraint, 0.0, np.inf, jac=jac_constr)
        )

        if self.signal_uncertainty_configuration.get("lambda", None) is not None:
            signal_lambda = self.signal_uncertainty_configuration["lambda"]

            def poiss_lamb(pars: np.ndarray) -> np.ndarray:
                """combined lambda expression"""
                return lam(pars) + signal_lambda(pars)

        else:
            poiss_lamb = lam

        self._main_model = MainModel(poiss_lamb)
        self._constraint_model = ConstraintModel(
            [
                {
                    "distribution_type": "multivariatenormal",
                    "args": [np.zeros(len(self.data)), corr],
                    "kwargs": {"domain": slice(1, len(B) + 1)},
                }
            ]
            + self.signal_uncertainty_configuration.get("constraint", [])
        )


class EffectiveSigma(DefaultPDFBase):
    r"""
    Simplified likelihood interface with variable Gaussian.
    Variable Gaussian has been inspired by :xref:`physics/0406120` sec. 3.6. This method
    modifies the effective :math:`B:=\sigma_{eff}` term in the Poisson distribution of the
    simplified likelihood framework. Note that this approach does not modify the Gaussian
    of the likelihood, the naming of the approach is purely because it is originated from
    :xref:`physics/0406120` sec. 3.6. The effective sigma term of the Poissonian can be
    modified using upper, :math:`\sigma^+` and lower :math:`\sigma^-` envelops of the
    absolute background uncertainties (see eqs 18-19 in :xref:`physics/0406120`).

    .. math::

        \sigma_{eff}(\theta) = \sqrt{\sigma^+\sigma^-  + (\sigma^+ - \sigma^-)(\theta - n_{bkg})}

    where the simplified likelihoo is modified as

    .. math::

        \mathcal{L}(\mu,\theta) = \left[\prod_i^N{\rm Poiss}(n^i_{obs}|\mu n^i_s + n^i_{bkg} +
        \theta^i\sigma_{eff}^i(\theta)) \right]\cdot \mathcal{N}(\theta| 0, \rho)

    .. note::

        This likelihood is constrained by

        .. math::

            n^i_{bkg} + \theta^i\sigma_{eff}^i(\theta) \geq 0

    Args:
        signal_yields (``np.ndarray``): signal yields
        background_yields (``np.ndarray``): background yields
        data (``np.ndarray``): observations
        correlation_matrix (``np.ndarray``): correlations between regions
        absolute_uncertainty_envelops (``List[Tuple[float, float]]``): upper and lower uncertainty
          envelops for each background yield.
        signal_uncertainty_configuration (``Dict[Text, Any]]``, default ``None``): Configuration
          input for signal uncertainties

          * absolute_uncertainties (``List[float]``): Absolute uncertainties for the signal
          * absolute_uncertainty_envelops (``List[Tuple[float, float]]``): upper and lower
              uncertainty envelops
          * correlation_matrix (``List[List[float]]``): Correlation matrix
          * third_moments (``List[float]``): diagonal elemetns of the third moment
    """

    name: Text = "default_pdf.effective_sigma"
    """Name of the backend"""
    version: Text = DefaultPDFBase.version
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = DefaultPDFBase.spey_requires
    """Spey version required for the backend"""
    doi: List[Text] = ["10.1142/9781860948985_0013"]
    """Citable DOI for the backend"""
    arXiv: List[Text] = ["physics/0406120"]
    """arXiv reference for the backend"""

    def __init__(
        self,
        signal_yields: np.ndarray,
        background_yields: np.ndarray,
        data: np.ndarray,
        correlation_matrix: np.ndarray,
        absolute_uncertainty_envelops: List[Tuple[float, float]],
        signal_uncertainty_configuration: Optional[Dict[Text, Any]] = None,
    ):
        assert len(absolute_uncertainty_envelops) == len(
            background_yields
        ), "Dimensionality of the uncertainty envelops does not match to the number of regions."
        assert len(correlation_matrix) == len(
            background_yields
        ), "Dimensionality of the correlation matrix does not match to the number of regions."

        sigma_plus, sigma_minus = [], []
        for upper, lower in absolute_uncertainty_envelops:
            sigma_plus.append(abs(upper))
            sigma_minus.append(abs(lower))
        sigma_plus = np.array(sigma_plus)
        sigma_minus = np.array(sigma_minus)
        correlation_matrix = np.array(correlation_matrix)

        super().__init__(
            signal_yields=signal_yields,
            background_yields=background_yields,
            data=data,
            signal_uncertainty_configuration=signal_uncertainty_configuration,
        )

        self._constraint_model: ConstraintModel = ConstraintModel(
            [
                {
                    "distribution_type": "multivariatenormal",
                    "args": [np.zeros(len(self.data)), correlation_matrix],
                    "kwargs": {"domain": slice(1, None)},
                }
            ]
            + self.signal_uncertainty_configuration.get("constraint", [])
        )

        A = self.background_yields

        # arXiv:pyhsics/0406120 eq. 18-19
        def effective_sigma(pars: np.ndarray) -> np.ndarray:
            """Compute effective sigma"""
            # clip from 1e-10 to avoid negative or zero values
            # this allows more numeric stability
            return np.sqrt(
                np.clip(
                    sigma_plus * sigma_minus
                    + (sigma_plus - sigma_minus) * (pars[slice(1, len(A) + 1)] - A),
                    1e-10,
                    None,
                )
            )

        def lam(pars: np.ndarray) -> np.ndarray:
            """Compute lambda for Main model"""
            return (
                A
                + effective_sigma(pars) * pars[slice(1, len(A) + 1)]
                + pars[0] * self.signal_yields
            )

        def constraint(pars: np.ndarray) -> np.ndarray:
            """Compute the constraint term"""
            return A + effective_sigma(pars) * pars[slice(1, len(A) + 1)]

        jac_constr = jacobian(constraint)
        self.constraints.append(
            NonlinearConstraint(constraint, 0.0, np.inf, jac=jac_constr)
        )
        if self.signal_uncertainty_configuration.get("lambda", None) is not None:
            signal_lambda = self.signal_uncertainty_configuration["lambda"]

            def poiss_lamb(pars: np.ndarray) -> np.ndarray:
                """combined lambda expression"""
                return lam(pars) + signal_lambda(pars)

        else:
            poiss_lamb = lam

        self._main_model = MainModel(poiss_lamb)



from enum import Enum, auto

__all__ = ["ExpectationType"]

# pylint: disable=C0103


class ExpectationType(Enum):
    """
    Expectation type has been used to determine the nature of the statistical model through out the package.
    It consists of three main arguments:

        * :obj:`observed` : indicates that the fit of the statistical model will be done over experimental data
        * :obj:`aposteriori`: as in :obj:`observed` the fit will be done over data where the likelihood results will
          be identical to :obj:`observed`, computation of :math:`CL_s` values will be done for by centralising the test
          statistics around background.
        * :obj:`apriori`: theorists are generatly interested in difference of their model from the SM simulation. Hence
          this option will overwrite the observed data in the statistical model with simulated background values and performs
          the computation with respect to prefit values, meaning prior to the experimental observation. :math:`CL_s` values
          are again computed by centralising the test statistics around the background i.e. SM background.

    User can simply set the value of :obj:`expected` to a desired :obj:`ExpectationType`:

    .. code-block:: python3

        >>> expected = spey.ExpectationType.aposteriori

    This will trigger appropriate action to be taken through out the package.
    """

    apriori = auto()
    aposteriori = auto()
    observed = auto()

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        current = str(self)
        if isinstance(other, ExpectationType):
            other = str(other)
            return current == other
        if isinstance(other, str):
            current = str(self)
            return other == current
        if isinstance(other, bool):
            return self == (
                ExpectationType.apriori if other else ExpectationType.observed
            )
        if other is None:
            return False

        raise ValueError(f"Unknown comparison: type({other}) = {type(other)}")



"""Abstract Methods for backend objects"""

from abc import ABC, abstractmethod
from typing import Callable, List, Optional, Text, Tuple, Union

import numpy as np

from spey.base.model_config import ModelConfig
from spey.utils import ExpectationType

__all__ = ["BackendBase"]


def __dir__():
    return __all__


class BackendBase(ABC):
    """
    An abstract class construction to enforce certain behaviour on statistical model backend.
    In order to perform certain computations, ``spey`` needs to have access to specific
    function constructions such as precsription to form likelihood. Hence, each backend is
    required to inherit :obj:`~spey.BackendBase`.
    """

    @property
    def is_alive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        return True

    @abstractmethod
    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: float = 10.0
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``10.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.model_config.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """

    @abstractmethod
    def get_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[Union[List[float], np.ndarray]] = None,
    ) -> Callable[[np.ndarray], float]:
        r"""
        Generate function to compute :math:`\log\mathcal{L}(\mu, \theta)` where :math:`\mu` is the
        parameter of interest and :math:`\theta` are nuisance parameters.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and computes
            :math:`\log\mathcal{L}(\mu, \theta)`.
        """

    def get_objective_function(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[Union[List[float], np.ndarray]] = None,
        do_grad: bool = True,
    ) -> Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]:
        r"""
        Objective function is the function to perform the optimisation on. This function is
        expected to be negative log-likelihood, :math:`-\log\mathcal{L}(\mu, \theta)`.
        Additionally, if available it canbe bundled with the gradient of negative log-likelihood.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit
            do_grad (``bool``, default ``True``): If ``True`` return objective and its gradient
              as ``tuple`` (subject to availablility) if ``False`` only returns objective function.

        Returns:
            ``Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]``:
            Function which takes fit parameters (:math:`\mu` and :math:`\theta`) and returns either
            objective or objective and its gradient.
        """
        if do_grad:
            raise NotImplementedError("Gradient is not implemented by default.")

        logpdf = self.get_logpdf_func(expected=expected, data=data)
        return lambda pars: -logpdf(pars)

    def get_hessian_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[Union[List[float], np.ndarray]] = None,
    ) -> Callable[[np.ndarray], float]:
        r"""
        Currently Hessian of :math:`\log\mathcal{L}(\mu, \theta)` is only used to compute
        variance on :math:`\mu`. This method returns a callable function which takes fit
        parameters (:math:`\mu` and :math:`\theta`) and returns Hessian.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit

        Raises:
            :obj:`NotImplementedError`: If the Hessian of the backend has not been implemented.

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and
            returns Hessian of :math:`\log\mathcal{L}(\mu, \theta)`.
        """
        raise NotImplementedError("This method has not been implemented")

    def get_sampler(self, pars: np.ndarray) -> Callable[[int], np.ndarray]:
        r"""
        Retreives the function to sample from.

        Args:
            pars (:obj:`np.ndarray`): fit parameters (:math:`\mu` and :math:`\theta`)

        Raises:
            :obj:`NotImplementedError`: If the sampler for the backend has not been implemented.

        Returns:
            ``Callable[[int], np.ndarray]``:
            Function that takes ``number_of_samples`` as input and draws as many samples
            from the statistical model.
        """
        raise NotImplementedError("This method has not been implemented")

    def expected_data(self, pars: List[float]) -> List[float]:
        r"""
        Compute the expected value of the statistical model. This function is mainly used to
        generate Asimov data within the package, see
        :func:`~spey.StatisticalModel.generate_asimov_data`.

        Args:
            pars (``List[float]``): nuisance, :math:`\theta` and parameter of interest,
              :math:`\mu`.

        Returns:
            ``List[float]``:
            Expected data of the statistical model
        """
        raise NotImplementedError("This method has not been implemented")

    def combine(self, other, **kwargs):
        """
        A routine to combine to statistical models.

        .. note::

            This function is only available if the backend has a specific routine
            for combination between same or other backends.

        Args:
            other (:obj:`~spey.BackendBase`): Statistical model object to be combined.

        Raises:
            ``NotImplementedError``: If the backend does not have a combination scheme.

        Returns:
            :obj:`~spey.BackendBase`:
            Create a new statistical model from combination of this and other one.
        """
        raise NotImplementedError("This method does not have combination implementation.")

    def negative_loglikelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        **kwargs,
    ) -> Tuple[float, np.ndarray]:
        r"""
        Backend specific method to compute negative log-likelihood for a parameter of interest
        :math:`\mu`.

        .. note::

            Interface first calls backend specific methods to compute likelihood. If they are not
            implemented, it optimizes objective function through ``spey`` interface. Either prescription
            to optimizing the likelihood or objective function must be available for a backend to
            be sucessfully integrated to the ``spey`` interface.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            kwargs: keyword arguments for the optimiser.

        Raises:
            :obj:`NotImplementedError`: If the method is not available for the backend.

        Returns:
            ``Tuple[float, np.ndarray]``:
            value of negative log-likelihood at POI of interest and fit parameters
            (:math:`\mu` and :math:`\theta`).
        """
        raise NotImplementedError("This method has not been implemented")

    def asimov_negative_loglikelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        test_statistics: Text = "qtilde",
        **kwargs,
    ) -> Tuple[float, np.ndarray]:
        r"""
        Compute negative log-likelihood at fixed :math:`\mu` for Asimov data.

        .. note::

            Interface first calls backend specific methods to compute likelihood. If they are not
            implemented, it optimizes objective function through ``spey`` interface. Either prescription
            to optimizing the likelihood or objective function must be available for a backend to
            be sucessfully integrated to the ``spey`` interface.

        Args:
            poi_test (``float``, default ``1.0``\): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistics (``Text``, default ``"qtilde"``): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence ``allow_negative_signal``
                    assumed to be ``False``. If this function has been executed by user, ``spey``
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through ``spey`` this
                    is taken care of automatically in the backend.

              * ``'q'``\: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``\: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            kwargs: keyword arguments for the optimiser.

        Raises:
            :obj:`NotImplementedError`: If the method is not available for the backend.

        Returns:
            ``Tuple[float, np.ndarray]``\:
            value of negative log-likelihood at POI of interest and fit parameters
            (:math:`\mu` and :math:`\theta`).
        """
        raise NotImplementedError("This method has not been implemented")

    def minimize_negative_loglikelihood(
        self,
        expected: ExpectationType = ExpectationType.observed,
        allow_negative_signal: bool = True,
        **kwargs,
    ) -> Tuple[float, np.ndarray]:
        r"""
        A backend specific method to minimize negative log-likelihood.

        .. note::

            Interface first calls backend specific methods to compute likelihood. If they are not
            implemented, it optimizes objective function through ``spey`` interface. Either prescription
            to optimizing the likelihood or objective function must be available for a backend to
            be sucessfully integrated to the ``spey`` interface.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            kwargs: keyword arguments for the optimiser.

        Raises:
            :obj:`NotImplementedError`: If the method is not available for the backend.

        Returns:
            ``Tuple[float, np.ndarray]``:
            value of negative log-likelihood and fit parameters (:math:`\mu` and :math:`\theta`).
        """
        raise NotImplementedError("This method has not been implemented")

    def minimize_asimov_negative_loglikelihood(
        self,
        expected: ExpectationType = ExpectationType.observed,
        test_statistics: Text = "qtilde",
        **kwargs,
    ) -> Tuple[float, np.ndarray]:
        r"""
        A backend specific method to minimize negative log-likelihood for Asimov data.

        .. note::

            Interface first calls backend specific methods to compute likelihood. If they are not
            implemented, it optimizes objective function through ``spey`` interface. Either prescription
            to optimizing the likelihood or objective function must be available for a backend to
            be sucessfully integrated to the ``spey`` interface.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistics (``Text``, default ``"qtilde"``): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence ``allow_negative_signal``
                    assumed to be ``False``. If this function has been executed by user, ``spey``
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through ``spey`` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            kwargs: keyword arguments for the optimiser.

        Raises:
            :obj:`NotImplementedError`: If the method is not available for the backend.

        Returns:
            ``Tuple[float, np.ndarray]``:
            value of negative log-likelihood and fit parameters (:math:`\mu` and :math:`\theta`).
        """
        raise NotImplementedError("This method has not been implemented")


class ConverterBase(ABC):
    """
    An abstract class construction to enforce certain behaviour on statistical model backend.
    This base class is used to act as a midle function where the function can act as a converter
    between two statistical models. It has to have a call function which needs to return a
    :obj:`~spey.BackendBase` object.

    .. note::

        ``ConverterBase`` object is not expected to have an ``__init__`` method that expect
        arguments or keyword arguments. This function should have a ``__call__`` method that
        returns :obj:`~spey.BackendBase` object.

    Example:

    .. code:: python3

        >>> class MyStatConverter(ConverterBase):
        >>>     name= "example.converter"
        >>>     version="0.0.1"
        >>>     author="Tom Bombadil"
        >>>     spey_requires=">0.1.0"

        >>>     def __call__(self, stat_model: spey.StatisticalModel) -> BackendBase:
        >>>         # do something
        >>>         return UncorrelatedBackground(...)
    """

    def __call__(self, *args, **kwargs) -> BackendBase:
        """
        Function that compiles the :obj:`~spey.BackendBase` object.

        Raises:
            ``NotImplementedError``: If ``__call__`` function is not implemented for the class.

        Returns:
            :obj:`~spey.BackendBase`:
            This function should return any :obj:`~spey.BackendBase` object.
        """
        raise NotImplementedError("Invalid implementation of ConverterBase object")



"""Functions for computation of test statistic"""

import logging
import warnings
from typing import Callable, Text, Tuple

import numpy as np

from spey.system.exceptions import UnknownTestStatistics, AsimovTestStatZero

__all__ = ["qmu", "qmu_tilde", "q0", "get_test_statistic", "compute_teststatistics"]

log = logging.getLogger("Spey")

# pylint: disable=W1203


def qmu_tilde(
    mu: float, muhat: float, max_logpdf: float, logpdf: Callable[[float], float]
) -> float:
    r"""
    Alternative test statistics, :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`.

    .. math::

        \tilde{q}_{\mu} = \begin{cases}
            0 & \text{if}\ \hat{\mu} > \mu\ , \\
            -2\log\left( \frac{\mathcal{L}(\mu, \theta_\mu)}{\mathcal{L}(\max(\hat{\mu}, 0), \hat{\theta})} \right) & \text{otherwise}
        \end{cases}

    .. warning::

        Note that this assumes that :math:`\hat\mu\geq0`, hence :obj:`allow_negative_signal`
        assumed to be ``False``. If this function has been executed by user, :obj:`spey`
        assumes that this is taken care of throughout the external code consistently.
        Whilst computing p-values or upper limit on :math:`\mu` through :obj:`spey` this
        is taken care of automatically in the backend.

    Args:
        mu (``float``): parameter of interest, :math:`\mu`.
        muhat (``float``): :math:`\hat\mu` value that maximizes the likelihood.
        max_logpdf (``float``): maximum value of :math:`\log\mathcal{L}`.
        logpdf (``Callable[[float], float]``): :math:`\log\mathcal{L}(\mu, \theta_\mu)`.

    Returns:
        ``float``:
        the value of :math:`\tilde{q}_{\mu}`.
    """
    # NOTE Comparing muhat to mu is important in the cases where obs >> bkg.
    # it can lead sig << bkg scenario to be excluded which should not happen
    return (
        0.0
        if muhat > mu
        else np.clip(
            -2.0 * (logpdf(mu) - (max_logpdf if muhat >= 0.0 else logpdf(0.0))), 0.0, None
        )
    )


def qmu(
    mu: float, muhat: float, max_logpdf: float, logpdf: Callable[[float], float]
) -> float:
    r"""
    Test statistic :math:`q_{\mu}`, see eq. (54) of :xref:`1007.1727`
    
    .. math::

        q_{\mu} = \begin{cases}
            0 & \text{if}\ \hat{\mu} > \mu\ ,\\
            -2\log\left( \frac{\mathcal{L}(\mu, \theta_\mu)}{\mathcal{L}(\hat{\mu}, \hat{\theta})} \right) & \text{otherwise}
        \end{cases}

    Args:
        mu (``float``): parameter of interest, :math:`\mu`.
        muhat (``float``): :math:`\hat\mu` value that maximizes the likelihood.
        max_logpdf (``float``): maximum value of :math:`\log\mathcal{L}`.
        logpdf (``Callable[[float], float]``): :math:`\log\mathcal{L}(\mu, \theta_\mu)`.

    Returns:
        ``float``:
        the value of :math:`q_{\mu}`.
    """
    return 0.0 if muhat > mu else np.clip(-2.0 * (logpdf(mu) - max_logpdf), 0.0, None)


def q0(
    mu: float, muhat: float, max_logpdf: float, logpdf: Callable[[float], float]
) -> float:
    r"""
    Discovery test statistics, :math:`q_{0}` see eq. (47) of :xref:`1007.1727`.

    .. math::

        q_0 = \begin{cases}
            0 & \text{if}\ \hat{\mu} < 0\ ,\\
            -2\log\left( \frac{\mathcal{L}(0, \theta_0)}{\mathcal{L}(\hat{\mu}, \hat{\theta})} \right) & \text{otherwise}
        \end{cases}

    Args:
        mu (``float``): parameter of interest, :math:`\mu`.

          .. note::

            ``mu`` argument is overwritten by zero. This input is only for consistency with
            other test statistic functions.

        muhat (``float``): :math:`\hat\mu` value that maximizes the likelihood.
        max_logpdf (``float``): maximum value of :math:`\log\mathcal{L}`.
        logpdf (``Callable[[float], float]``): :math:`\log\mathcal{L}(\mu, \theta_\mu)`.

    Returns:
        ``float``:
        the value of :math:`q_{0}`.
    """
    mu = 0.0
    return 0.0 if muhat < 0.0 else np.clip(-2.0 * (logpdf(mu) - max_logpdf), 0.0, None)


def get_test_statistic(
    test_stat: Text,
) -> Callable[[float, float, float, Callable[[float], float]], float]:
    r"""
    Retreive the test statistic function

    Args:
        test_stat (``Text``): test statistic.

          * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
            :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
            (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

            .. warning::

                Note that this assumes that :math:`\hat\mu\geq0`, hence :obj:`allow_negative_signal`
                assumed to be ``False``. If this function has been executed by user, :obj:`spey`
                assumes that this is taken care of throughout the external code consistently.
                Whilst computing p-values or upper limit on :math:`\mu` through :obj:`spey` this
                is taken care of automatically in the backend.

          * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
            eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
          * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
            of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

    Raises:
        :obj:`~spey.system.exceptions.UnknownTestStatistics`: If the ``test_stat`` input does not match
          any of the above.

    Returns:
        ``Callable[[float, float, float, Callable[[float], float]], float]``:
        returns the function to compute test statistic
    """
    # syntax with mu is not necessary so accomodate both
    if test_stat in ["qmu", "q"]:
        test_stat = "q"
    elif test_stat in ["qtilde", "qmutilde"]:
        test_stat = "qmutilde"
    options = {"qmutilde": qmu_tilde, "q": qmu, "q0": q0}

    if options.get(test_stat, False) is False:
        raise UnknownTestStatistics(
            f"Requested test statistics {test_stat} does not exist."
        )

    return options[test_stat]


def compute_teststatistics(
    mu: float,
    maximum_likelihood: Tuple[float, float],
    logpdf: Callable[[float], float],
    maximum_asimov_likelihood: Tuple[float, float],
    asimov_logpdf: Callable[[float], float],
    teststat: Text,
) -> Tuple[float, float, float]:
    r"""
    Compute test statistics

    Args:
        mu (``float``): parameter of interest, :math:`\mu`.
        maximum_likelihood (``Tuple[float, float]``): (:math:`\hat\mu` and :math:`\arg\min\log\mathcal{L}`)
        logpdf (``Callable[[float], float]``): function to compute :math:`\log\mathcal{L}` with fixed :math:`\mu`.
        maximum_asimov_likelihood (``Tuple[float, float]``): (:math:`\hat\mu_A` and :math:`\arg\min\log\mathcal{L}_A`)
        asimov_logpdf (``Callable[[float], float]``): function to compute :math:`\log\mathcal{L}_A` with fixed :math:`\mu`.
        teststat (``Text``): test statistic.

          * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
            :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
            (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

            .. warning::

                Note that this assumes that :math:`\hat\mu\geq0`, hence :obj:`allow_negative_signal`
                assumed to be ``False``. If this function has been executed by user, :obj:`spey`
                assumes that this is taken care of throughout the external code consistently.
                Whilst computing p-values or upper limit on :math:`\mu` through :obj:`spey` this
                is taken care of automatically in the backend.

          * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
            eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
          * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
            of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

    Raises:
        :obj:`~spey.system.exceptions.AsimovTestStatZero`: Raised if Asimov test statistic is zero.

    Returns:
        ``Tuple[float, float, float]``:
        :math:`\sqrt{q_\mu}`, :math:`\sqrt{q_{\mu,A}}` and :math:`\Delta(\sqrt{q_\mu}, \sqrt{q_{\mu,A}})`
    """
    teststat_func = get_test_statistic(teststat)

    muhat, min_nll = maximum_likelihood
    muhatA, min_nllA = maximum_asimov_likelihood

    # max_logpdf = -min_nll
    qmu_ = teststat_func(mu, muhat, -min_nll, logpdf)
    qmuA = teststat_func(mu, muhatA, -min_nllA, asimov_logpdf)
    sqrt_qmu, sqrt_qmuA = np.sqrt(qmu_), np.sqrt(qmuA)

    if teststat in ["q", "q0", "qmu"]:
        delta_teststat = sqrt_qmu - sqrt_qmuA
    else:
        # arXiv:1007.1727 eq. 66
        if sqrt_qmu <= sqrt_qmuA:
            delta_teststat = sqrt_qmu - sqrt_qmuA
        else:
            with warnings.catch_warnings(record=True):
                delta_teststat = np.true_divide(qmu_ - qmuA, 2.0 * sqrt_qmuA)
            if sqrt_qmuA == 0:
                raise AsimovTestStatZero()
    log.debug(
        f"sqrt_qmu = {sqrt_qmu}, sqrt_qmuA = {sqrt_qmuA}, delta_teststat = {delta_teststat}"
    )
    return sqrt_qmu, sqrt_qmuA, delta_teststat



from typing import Callable, Optional, Tuple, List

from autograd import numpy as np

from .interface.statistical_model import StatisticalModel
from .utils import ExpectationType

# pylint: disable=E1101

__all__ = ["value_and_grad", "hessian"]


def __dir__():
    return __all__


def value_and_grad(
    statistical_model: StatisticalModel,
    expected: ExpectationType = ExpectationType.observed,
    data: Optional[List[float]] = None,
) -> Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]]:
    """
    Retreive function to compute negative log-likelihood and its gradient.

    .. versionadded:: 0.1.6

    Args:
        statistical_model (~spey.StatisticalModel): statistical model to be used.
        expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
          p-values to be computed.

          * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
            prescriotion which means that the experimental data will be assumed to be the truth
            (default).
          * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
            post-fit prescriotion which means that the experimental data will be assumed to be
            the truth.
          * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
            prescription which means that the SM will be assumed to be the truth.

        data (``List[float]``, default ``None``): input data that to fit. If `None` observed
          data will be used.

    Returns:
        ``Callable[[np.ndarray], Tuple[np.ndarray, np.ndarray]]``:
        negative log-likelihood and its gradient with respect to nuisance parameters
    """
    val_and_grad = statistical_model.backend.get_objective_function(
        expected=expected, data=None if data is None else np.array(data), do_grad=True
    )
    return lambda pars: val_and_grad(np.array(pars))


def hessian(
    statistical_model: StatisticalModel,
    expected: ExpectationType = ExpectationType.observed,
    data: Optional[List[float]] = None,
) -> Callable[[np.ndarray], np.ndarray]:
    r"""
    Retreive the function to compute Hessian of negative log-likelihood

    .. math::

        {\rm Hessian} = -\frac{\partial^2\mathcal{L}(\theta)}{\partial\theta_i\partial\theta_j}

    .. versionadded:: 0.1.6

    Args:
        statistical_model (~spey.StatisticalModel): statistical model to be used.
        expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
          p-values to be computed.

          * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
            prescriotion which means that the experimental data will be assumed to be the truth
            (default).
          * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
            post-fit prescriotion which means that the experimental data will be assumed to be
            the truth.
          * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
            prescription which means that the SM will be assumed to be the truth.

        data (``List[float]``, default ``None``): input data that to fit. If `None` observed
          data will be used.

    Returns:
        ``Callable[[np.ndarray], np.ndarray]``:
        function to compute hessian of negative log-likelihood
    """
    hess = statistical_model.backend.get_hessian_logpdf_func(
        expected=expected, data=None if data is None else np.array(data)
    )
    return lambda pars: -1.0 * hess(np.array(pars))



from typing import Any, Dict, List, Text, Tuple

import autograd.numpy as np

from spey.helper_functions import correlation_to_covariance

from .third_moment import third_moment_expansion

# pylint: disable=E1101,E1120


def constraint_from_corr(
    correlation_matrix: List[List[float]], size: int, domain: slice
) -> List[Dict[Text, Any]]:
    """
    Derive constraints from inputs

    Args:
        correlation_matrix (``List[List[float]]``): correlation matrix
        size (``int``): size of the signal vector
        domain (``slice``): domain of the nuisances

    Returns:
        ``List[Dict[Text, Any]]``:
    """
    if correlation_matrix is not None:
        corr = np.array(correlation_matrix)
        constraint_term = [
            {
                "distribution_type": "multivariatenormal",
                "args": [np.zeros(size), corr],
                "kwargs": {"domain": domain},
            }
        ]
    else:
        constraint_term = [
            {
                "distribution_type": "normal",
                "args": [np.zeros(size), np.ones(size)],
                "kwargs": {"domain": domain},
            }
        ]

    return constraint_term


def signal_uncertainty_synthesizer(
    signal_yields: List[float],
    absolute_uncertainties: List[float] = None,
    absolute_uncertainty_envelops: List[Tuple[float, float]] = None,
    correlation_matrix: List[List[float]] = None,
    third_moments: List[float] = None,
    domain: slice = None,
) -> Dict[Text, np.ndarray]:
    """
    Synthesize signal uncertainties

    Args:
        signal_yields (``List[float]``): signal yields
        absolute_uncertainties (``List[float]``, default ``None``): absolute uncertainties
        absolute_uncertainty_envelops (``List[Tuple[float, float]]``, default ``None``):
            asymmetric uncertainty envelops (upper, lower)
        correlation_matrix (``List[List[float]]``, default ``None``): correlation matrix
        third_moments (``List[float]``, default ``None``): third moments
        domain (``slice``, default ``None``): domain of the nuisances

    Raises:
        ``ValueError``: if inconsistent number of input has been provided

    Returns:
        ``Dict[Text, np.ndarray]``:
        Inputs for the main model and the constraint term
    """
    assert domain is not None, "Invalid domain"
    signal_yields = np.array(signal_yields)

    if absolute_uncertainties is not None and third_moments is None:
        absolute_uncertainties = np.array(absolute_uncertainties)

        def lam_signal(pars: np.ndarray) -> np.ndarray:
            return absolute_uncertainties * pars[domain]

        constraint_term = constraint_from_corr(
            correlation_matrix, len(absolute_uncertainties), domain
        )

    elif absolute_uncertainty_envelops is not None:
        sigma_plus, sigma_minus = [], []
        for upper, lower in absolute_uncertainty_envelops:
            sigma_plus.append(abs(upper))
            sigma_minus.append(abs(lower))
        sigma_plus = np.array(sigma_plus)
        sigma_minus = np.array(sigma_minus)

        # arXiv:pyhsics/0406120 eq. 18-19
        def effective_sigma(pars: np.ndarray) -> np.ndarray:
            """Compute effective sigma"""
            return np.sqrt(
                sigma_plus * sigma_minus
                + (sigma_plus - sigma_minus) * (pars[domain] - signal_yields)
            )

        def lam_signal(pars: np.ndarray) -> np.ndarray:
            """Compute lambda for Main model"""
            return effective_sigma(pars) * pars[domain]

        constraint_term = constraint_from_corr(
            correlation_matrix, len(sigma_plus), domain
        )

    elif all(
        x is not None for x in [third_moments, correlation_matrix, absolute_uncertainties]
    ):
        correlation_matrix = np.array(correlation_matrix)
        absolute_uncertainties = np.array(absolute_uncertainties)
        third_moments = np.array(third_moments)
        cov = correlation_to_covariance(correlation_matrix, absolute_uncertainties)

        A, B, C, corr = third_moment_expansion(signal_yields, cov, third_moments, True)

        def lam_signal(pars: np.ndarray) -> np.ndarray:
            """
            Compute lambda for Main model with third moment expansion.
            For details see above eq 2.6 in :xref:`1809.05548`

            Args:
                pars (``np.ndarray``): nuisance parameters

            Returns:
                ``np.ndarray``:
                expectation value of the poisson distribution with respect to
                nuisance parameters.
            """
            return A + B * pars[domain] + C * np.square(pars[domain])

        constraint_term = [
            {
                "distribution_type": "multivariatenormal",
                "args": [np.zeros(len(signal_yields)), corr],
                "kwargs": {"domain": domain},
            }
        ]

    else:
        raise ValueError("Inconsistent input.")

    return {"lambda": lam_signal, "constraint": constraint_term}



from .backend_base import BackendBase, ConverterBase
from .hypotest_base import HypothesisTestingBase
from .model_config import ModelConfig
from .recorder import Recorder

__all__ = [
    "BackendBase",
    "HypothesisTestingBase",
    "ModelConfig",
    "Recorder",
    "ConverterBase",
]



"""
Routines for computing pvalues

Similar implementation can be found in https://github.com/scikit-hep/pyhf/blob/main/src/pyhf/infer/calculators.py
"""

import warnings
from typing import List, Tuple, Union

import numpy as np

from .distributions import (
    AsymptoticTestStatisticsDistribution,
    EmpricTestStatisticsDistribution,
)

__all__ = ["pvalues", "expected_pvalues"]


def __dir__():
    return __all__


def pvalues(
    delta_test_statistic: float,
    sig_plus_bkg_distribution: Union[
        AsymptoticTestStatisticsDistribution, EmpricTestStatisticsDistribution
    ],
    bkg_only_distribution: Union[
        AsymptoticTestStatisticsDistribution, EmpricTestStatisticsDistribution
    ],
) -> Tuple[float, float, float]:
    r"""
    Calculate the p-values for the observed test statistic under the
    signal + background and background-only model hypotheses.

    .. math::
    
        p_{s+b}&=& \int_{-\infty}^{-\sqrt{q_{\mu,A}} - \Delta q_\mu} \mathcal{N}(x| 0, 1) dx \\
        p_{b}&=& \int_{-\infty}^{-\Delta q_\mu} \mathcal{N}(x| 0, 1) dx \\
        p_{s} &=& p_{s+b}/ p_{b}

    where :math:`q_\mu` stands for the test statistic and A stands for Assimov.

    Args:
        delta_test_statistic (``float``): :math:`\Delta q_\mu`
        sig_plus_bkg_distribution (~spey.hypothesis_testing.asymptotic_calculator.AsymptoticTestStatisticsDistribution):
          the distribution for the signal + background hypothesis.
        bkg_only_distribution (~spey.hypothesis_testing.asymptotic_calculator.AsymptoticTestStatisticsDistribution):
          The distribution for the background-only hypothesis.

    Returns:
        ``Tuple[float, float, float]``:
        The p-values for the test statistic corresponding to the :math:`CL_{s+b}`,
        :math:`CL_{b}`, and :math:`CL_{s}`.
        
    .. seealso::
    
        :func:`~spey.hypothesis_testing.test_statistics.compute_teststatistics`, 
        :func:`~spey.hypothesis_testing.asymptotic_calculator.compute_asymptotic_confidence_level`,
        :func:`~spey.hypothesis_testing.utils.expected_pvalues`
    """
    CLsb = sig_plus_bkg_distribution.pvalue(delta_test_statistic)
    CLb = bkg_only_distribution.pvalue(delta_test_statistic)
    with warnings.catch_warnings(record=True):
        CLs = np.true_divide(CLsb, CLb, dtype=np.float64)
    return CLsb, CLb, CLs if CLb != 0.0 else 0.0


def expected_pvalues(
    sig_plus_bkg_distribution: Union[
        AsymptoticTestStatisticsDistribution, EmpricTestStatisticsDistribution
    ],
    bkg_only_distribution: Union[
        AsymptoticTestStatisticsDistribution, EmpricTestStatisticsDistribution
    ],
) -> List[List]:
    r"""
    Calculate the :math:`p` values corresponding to the
    median significance of variations of the signal strength from the
    background only hypothesis :math:`\mu=0` at :math:`(-2,-1,0,1,2)\sigma`.

    .. math::
    
        p_{s+b}&=& \int_{-\infty}^{-\sqrt{q_{\mu,A}} - N\sigma} \mathcal{N}(x| 0, 1) dx \\
        p_{b}&=& \int_{-\infty}^{-N\sigma} \mathcal{N}(x| 0, 1) dx \\
        p_{s} &=& p_{s+b}/ p_{b}

    where :math:`q_\mu` stands for the test statistic and A stands for Assimov. 
    :math:`N\sigma\in[-2,-1,0,1,2]`.

    Args:
        sig_plus_bkg_distribution (~spey.hypothesis_testing.asymptotic_calculator.AsymptoticTestStatisticsDistribution):
          The distribution for the signal + background hypothesis.
        bkg_only_distribution (~spey.hypothesis_testing.asymptotic_calculator.AsymptoticTestStatisticsDistribution):
          The distribution for the background-only hypothesis.

    Returns:
        ``List[List]``:
        The p-values for the test statistic corresponding to the :math:`CL_{s+b}`,
        :math:`CL_{b}`, and :math:`CL_{s}`.
        
    .. seealso::
    
        :func:`~spey.hypothesis_testing.test_statistics.compute_teststatistics`, 
        :func:`~spey.hypothesis_testing.asymptotic_calculator.compute_asymptotic_confidence_level`,
        :func:`~spey.hypothesis_testing.utils.pvalues`
    """
    return list(
        map(
            list,
            zip(
                *[
                    pvalues(
                        bkg_only_distribution.expected_value(nsigma),
                        sig_plus_bkg_distribution,
                        bkg_only_distribution,
                    )
                    for nsigma in [2.0, 1.0, 0.0, -1.0, -2.0]
                ]
            ),
        )
    )



"""Function to display details about the spey installation"""

import platform
import sys
from subprocess import check_output

import numpy
import scipy
import semantic_version
import tqdm
from pkg_resources import get_distribution, iter_entry_points


def about() -> None:
    """Prints the information regarding spey installation"""

    print(check_output([sys.executable, "-m", "pip", "show", "spey"]).decode())
    print(f"Platform info:            {platform.platform(aliased=True)}")
    print(
        f"Python version:           {sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}"
    )
    print(f"Numpy version:            {numpy.__version__}")
    print(f"Scipy version:            {scipy.__version__}")
    print(f"Autograd version:         {get_distribution('autograd').version}")
    print(f"tqdm version:             {tqdm.__version__}")
    print(f"semantic_version version: {semantic_version.__version__}")

    print("\nInstalled backend plug-ins:\n")

    shown = ["spey"]
    plugin_devices = iter_entry_points("spey.backend.plugins")
    for d in plugin_devices:
        print(f"- {d.name} ({d.dist.project_name}-{d.dist.version})")
        if d.dist.project_name not in shown:
            print(
                check_output(
                    [sys.executable, "-m", "pip", "show", d.dist.project_name]
                ).decode()
            )
            shown.append(d.dist.project_name)



from setuptools import setup, find_packages

with open("README.md", mode="r", encoding="utf-8") as f:
    long_description = f.read()

with open("src/spey/_version.py", mode="r", encoding="UTF-8") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

requirements = [
    "numpy>=1.21.6, <2.0.0",
    "scipy>=1.10.0",
    "autograd>=1.5",
    "semantic_version~=2.10",
    "tqdm>=4.64.0",
    "requests>=2.31.0",
    "setuptools",
]

backend_plugins = [
    "default_pdf.uncorrelated_background = spey.backends.default_pdf:UncorrelatedBackground",
    "default_pdf.correlated_background = spey.backends.default_pdf:CorrelatedBackground",
    "default_pdf.third_moment_expansion = spey.backends.default_pdf:ThirdMomentExpansion",
    "default_pdf.effective_sigma = spey.backends.default_pdf:EffectiveSigma",
    "default_pdf.poisson = spey.backends.default_pdf.simple_pdf:Poisson",
    "default_pdf.normal = spey.backends.default_pdf.simple_pdf:Gaussian",
    "default_pdf.multivariate_normal = spey.backends.default_pdf.simple_pdf:MultivariateNormal",
]

setup(
    name="spey",
    version=version,
    description=("Smooth inference for reinterpretation studies"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SpeysideHEP/spey",
    project_urls={
        "Bug Tracker": "https://github.com/SpeysideHEP/spey/issues",
        "Documentation": "https://spey.readthedocs.io",
        "Repository": "https://github.com/SpeysideHEP/spey",
        "Homepage": "https://github.com/SpeysideHEP/spey",
        "Download": f"https://github.com/SpeysideHEP/spey/archive/refs/tags/v{version}.tar.gz",
    },
    download_url=f"https://github.com/SpeysideHEP/spey/archive/refs/tags/v{version}.tar.gz",
    author="Jack Y. Araz",
    author_email=("jack.araz@stonybrook.edu"),
    license="MIT",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    entry_points={"spey.backend.plugins": backend_plugins},
    install_requires=requirements,
    python_requires=">=3.8, <=3.12.6",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Physics",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    extras_require={
        "dev": ["pytest>=7.1.2", "pytest-cov>=3.0.0", "twine>=3.7.1", "wheel>=0.37.1"],
    },
)



"""This file contains basic likelihood implementations"""

from typing import Callable, List, Optional, Text, Tuple, Union

from autograd import hessian, jacobian
from autograd import numpy as np
from autograd import value_and_grad
from scipy.optimize import NonlinearConstraint

from spey._version import __version__
from spey.backends.distributions import MainModel
from spey.base import BackendBase, ModelConfig
from spey.system.exceptions import InvalidInput
from spey.utils import ExpectationType

# pylint: disable=E1101,E1120,W0613


class SimplePDFBase(BackendBase):
    """Base structure for simple PDFs"""

    name: Text = "__simplepdf_base__"
    """Name of the backend"""
    version: Text = __version__
    """Version of the backend"""
    author: Text = "SpeysideHEP"
    """Author of the backend"""
    spey_requires: Text = __version__
    """Spey version required for the backend"""

    __slots__ = [
        "data",
        "signal_yields",
        "background_yields",
        "_main_model",
        "_main_kwargs",
        "_config",
    ]

    def __init__(
        self,
        signal_yields: List[float],
        background_yields: List[float],
        data: List[int],
    ):
        self.data = np.array(data, dtype=np.float64)
        self.signal_yields = np.array(signal_yields, dtype=np.float64)
        self.background_yields = np.array(background_yields, dtype=np.float64)
        self._main_model = None
        """main model"""
        self._main_kwargs = {}
        """Keyword arguments for main model"""
        self.constraints = []
        """Constraints to be used during optimisation process"""

        minimum_poi = -np.inf
        if self.is_alive:
            minimum_poi = -np.min(
                self.background_yields[self.signal_yields > 0.0]
                / self.signal_yields[self.signal_yields > 0.0]
            )

        self._config = ModelConfig(
            poi_index=0,
            minimum_poi=minimum_poi,
            suggested_init=[1.0],
            suggested_bounds=[(minimum_poi, 10)],
        )

    @property
    def is_alive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        return np.any(self.signal_yields > 0.0)

    def config(
        self, allow_negative_signal: bool = True, poi_upper_bound: float = 10.0
    ) -> ModelConfig:
        r"""
        Model configuration.

        Args:
            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            poi_upper_bound (``float``, default ``40.0``): upper bound for parameter
              of interest, :math:`\mu`.

        Returns:
            ~spey.base.ModelConfig:
            Model configuration. Information regarding the position of POI in
            parameter list, suggested input and bounds.
        """
        if allow_negative_signal and poi_upper_bound == 10.0:
            return self._config

        return ModelConfig(
            self._config.poi_index,
            self._config.minimum_poi,
            self._config.suggested_init,
            [(0, poi_upper_bound)],
        )

    @property
    def main_model(self) -> MainModel:
        """retreive the main model distribution"""
        if self._main_model is None:

            def lam(pars: np.ndarray) -> np.ndarray:
                """
                Compute lambda for Main model with third moment expansion.
                For details see above eq 2.6 in :xref:`1809.05548`

                Args:
                    pars (``np.ndarray``): nuisance parameters

                Returns:
                    ``np.ndarray``:
                    expectation value of the poisson distribution with respect to
                    nuisance parameters.
                """
                return pars[0] * self.signal_yields + self.background_yields

            self._main_model = MainModel(lam, **self._main_kwargs)

        return self._main_model

    def get_objective_function(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[np.ndarray] = None,
        do_grad: bool = True,
    ) -> Callable[[np.ndarray], Union[Tuple[float, np.ndarray], float]]:
        r"""
        Objective function i.e. negative log-likelihood, :math:`-\log\mathcal{L}(\mu, \theta)`

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``np.ndarray``, default ``None``): input data that to fit
            do_grad (``bool``, default ``True``): If ``True`` return objective and its gradient
              as ``tuple`` if ``False`` only returns objective function.

        Returns:
            ``Callable[[np.ndarray], Union[float, Tuple[float, np.ndarray]]]``:
            Function which takes fit parameters (:math:`\mu` and :math:`\theta`) and returns either
            objective or objective and its gradient.
        """
        current_data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        )
        data = current_data if data is None else data

        def negative_loglikelihood(pars: np.ndarray) -> np.ndarray:
            """Compute twice negative log-likelihood"""
            return -self.main_model.log_prob(pars, data)

        if do_grad:
            return value_and_grad(negative_loglikelihood, argnum=0)

        return negative_loglikelihood

    def get_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[np.array] = None,
    ) -> Callable[[np.ndarray, np.ndarray], float]:
        r"""
        Generate function to compute :math:`\log\mathcal{L}(\mu, \theta)` where :math:`\mu` is the
        parameter of interest and :math:`\theta` are nuisance parameters.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``np.array``, default ``None``): input data that to fit

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and computes
            :math:`\log\mathcal{L}(\mu, \theta)`.
        """
        current_data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        )
        data = current_data if data is None else data

        return lambda pars: self.main_model.log_prob(pars, data)

    def get_hessian_logpdf_func(
        self,
        expected: ExpectationType = ExpectationType.observed,
        data: Optional[np.ndarray] = None,
    ) -> Callable[[np.ndarray], float]:
        r"""
        Currently Hessian of :math:`\log\mathcal{L}(\mu, \theta)` is only used to compute
        variance on :math:`\mu`. This method returns a callable function which takes fit
        parameters (:math:`\mu` and :math:`\theta`) and returns Hessian.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            data (``np.ndarray``, default ``None``): input data that to fit

        Returns:
            ``Callable[[np.ndarray], float]``:
            Function that takes fit parameters (:math:`\mu` and :math:`\theta`) and
            returns Hessian of :math:`\log\mathcal{L}(\mu, \theta)`.
        """
        current_data = (
            self.background_yields if expected == ExpectationType.apriori else self.data
        )
        data = current_data if data is None else data

        def log_prob(pars: np.ndarray) -> np.ndarray:
            """Compute log-probability"""
            return self.main_model.log_prob(pars, data)

        return hessian(log_prob, argnum=0)

    def get_sampler(self, pars: np.ndarray) -> Callable[[int], np.ndarray]:
        r"""
        Retreives the function to sample from.

        Args:
            pars (``np.ndarray``): fit parameters (:math:`\mu` and :math:`\theta`)
            include_auxiliary (``bool``): wether or not to include auxiliary data
              coming from the constraint model.

        Returns:
            ``Callable[[int, bool], np.ndarray]``:
            Function that takes ``number_of_samples`` as input and draws as many samples
            from the statistical model.
        """

        def sampler(sample_size: int, *args, **kwargs) -> np.ndarray:
            """
            Fucntion to generate samples.

            Args:
                sample_size (``int``): number of samples to be generated.

            Returns:
                ``np.ndarray``:
                generated samples
            """
            return self.main_model.sample(pars, sample_size)

        return sampler

    def expected_data(self, pars: List[float], **kwargs) -> List[float]:
        r"""
        Compute the expected value of the statistical model

        Args:
            pars (``List[float]``): nuisance, :math:`\theta` and parameter of interest,

        Returns:
            ``List[float]``:
            Expected data of the statistical model
        """
        return self.main_model.expected_data(pars)


class Poisson(SimplePDFBase):
    r"""
    Poisson distribution without uncertainty implementation.

    .. math::

        \mathcal{L}(\mu) = \prod_{i\in{\rm bins}}{\rm Poiss}(n^i|\mu n_s^i + n_b^i)

    where :math:`n_{s,b}` are signal and background yields and :math:`n` are the observations.

    Args:
        signal_yields (``List[float]``): signal yields
        background_yields (``List[float]``): background yields
        data (``List[int]``): data
        absolute_uncertainties (``List[float]``, optional): background uncertainties to be added.
          :math:`\mu n_s^i + n_b^i + \theta^i\sigma^i`.
    """

    name: Text = "default_pdf.poisson"
    """Name of the backend"""
    version: Text = SimplePDFBase.version
    """Version of the backend"""
    author: Text = SimplePDFBase.author
    """Author of the backend"""
    spey_requires: Text = SimplePDFBase.spey_requires
    """Spey version required for the backend"""

    def __init__(
        self,
        signal_yields: List[float],
        background_yields: List[float],
        data: List[int],
        absolute_uncertainties: Optional[List[float]] = None,
    ):
        super().__init__(
            signal_yields=signal_yields, background_yields=background_yields, data=data
        )

        if absolute_uncertainties is not None:
            self.absolute_uncertainties = np.array(absolute_uncertainties)
            assert len(self.absolute_uncertainties) == len(
                self.data
            ), "Invalid input dimension."

            def lam(pars: np.ndarray) -> np.ndarray:
                return (
                    pars[0] * self.signal_yields
                    + self.background_yields
                    + pars[slice(1, len(self.data) + 1)] * self.absolute_uncertainties
                )

            self._main_model = MainModel(lam)

            self._config = ModelConfig(
                poi_index=0,
                minimum_poi=self._config.minimum_poi,
                suggested_init=[1.0] * (len(self.data) + 1),
                suggested_bounds=[(self._config.minimum_poi, 10)]
                + [(None, None)] * len(self.data),
            )

            def constraint(pars: np.ndarray) -> np.ndarray:
                """Compute the constraint term"""
                return (
                    self.background_yields
                    + pars[slice(1, len(self.data) + 1)] * self.background_yields
                )

            jac_constr = jacobian(constraint)

            self.constraints.append(
                NonlinearConstraint(constraint, 0.0, np.inf, jac=jac_constr)
            )


class Gaussian(SimplePDFBase):
    r"""
    Gaussian distribution for uncorrelated likelihoods.

    .. math::

        \mathcal{L}(\mu) = \prod_{i\in{\rm bins}} \frac{1}{\sigma^i \sqrt{2\pi}}
        \exp\left[-\frac{1}{2} \left(\frac{\mu n_s^i + n_b^i - n^i}{\sigma^i} \right)^2 \right]

    where :math:`n_{s,b}` are signal and background yields and :math:`n` are the observations.

    .. versionadded:: 0.1.9

    Args:
        signal_yields (``List[float]``): signal yields
        background_yields (``List[float]``): background yields
        data (``List[int]``): data
        absolute_uncertainties (``List[float]``): absolute uncertainties on the background
    """

    name: Text = "default_pdf.normal"
    """Name of the backend"""
    version: Text = SimplePDFBase.version
    """Version of the backend"""
    author: Text = SimplePDFBase.author
    """Author of the backend"""
    spey_requires: Text = SimplePDFBase.spey_requires
    """Spey version required for the backend"""

    __slots__ = ["absolute_uncertainties"]

    def __init__(
        self,
        signal_yields: List[float],
        background_yields: List[float],
        data: List[int],
        absolute_uncertainties: List[float],
    ):
        super().__init__(
            signal_yields=signal_yields, background_yields=background_yields, data=data
        )
        self.absolute_uncertainties = np.array(absolute_uncertainties, dtype=np.float64)
        """absolute uncertainties on the background"""
        self._main_kwargs = {"cov": self.absolute_uncertainties, "pdf_type": "gauss"}


class MultivariateNormal(SimplePDFBase):
    r"""
    Multivariate Gaussian distribution.

    .. math::

        \mathcal{L}(\mu) = \frac{1}{\sqrt{(2\pi)^k {\rm det}[\Sigma] }}
        \exp\left[-\frac{1}{2} (\mu n_s + n_b - n)\Sigma^{-1} (\mu n_s + n_b - n)^T \right]

    where :math:`n_{s,b}` are signal and background yields and :math:`n` are the observations.

    .. versionadded:: 0.1.9

    Args:
        signal_yields (``List[float]``): signal yields
        background_yields (``List[float]``): background yields
        data (``List[int]``): data
        covariance_matrix (``List[List[float]]``): covariance matrix (square matrix)

          * If you have correlation matrix and absolute uncertainties please use
            :func:`~spey.helper_functions.correlation_to_covariance`

    """

    name: Text = "default_pdf.multivariate_normal"
    """Name of the backend"""
    version: Text = SimplePDFBase.version
    """Version of the backend"""
    author: Text = SimplePDFBase.author
    """Author of the backend"""
    spey_requires: Text = SimplePDFBase.spey_requires
    """Spey version required for the backend"""

    __slots__ = ["covariance_matrix"]

    def __init__(
        self,
        signal_yields: List[float],
        background_yields: List[float],
        data: List[int],
        covariance_matrix: List[List[float]],
    ):
        super().__init__(
            signal_yields=signal_yields, background_yields=background_yields, data=data
        )
        self.covariance_matrix = np.array(covariance_matrix, dtype=np.float64)
        if (
            self.covariance_matrix.shape[0] != len(self.background_yields)
            and len(self.covariance_matrix.shape) == 2
        ):
            raise InvalidInput(
                "Dimensionality of the covariance matrix should match to the background"
            )

        self._main_kwargs = {
            "cov": self.covariance_matrix,
            "pdf_type": "multivariategauss",
        }



"""Tools for computing upper limit on parameter of interest"""

import logging
import warnings
from functools import partial
from typing import Callable, List, Tuple, Union, Literal

import numpy as np
import scipy

from spey.hypothesis_testing.test_statistics import compute_teststatistics
from spey.utils import ExpectationType

from .asymptotic_calculator import compute_asymptotic_confidence_level

__all__ = ["find_poi_upper_limit", "find_root_limits"]


def __dir__():
    return __all__


log = logging.getLogger("Spey")

# pylint: disable=W1203,C0103


class ComputerWrapper:
    """
    Wrapper for the computer function to track inputs and outputs

    Args:
        computer (``Callable[[float], float]``): desired function to be computed
    """

    def __init__(self, computer: Callable[[float], float]):
        self.computer = computer
        self._results = []

    def __call__(self, value: float) -> float:
        """Compute the input function and return its value"""
        self._results.append((value, self.computer(value)))
        return self[-1]

    def __getitem__(self, item: int) -> float:
        """Return result"""
        return self._results[item][1]

    def get_value(self, index: int) -> float:
        """
        Get input value of the execution

        Args:
            index (``int``): index of the execution

        Returns:
            ``float``:
            returns the input value of the execution
        """
        return self._results[index][0]


def find_root_limits(
    computer: Callable[[float], float],
    loc: float = 0.0,
    low_ini: float = 1.0,
    hig_ini: float = 1.0,
    low_bound: float = 1e-10,
    hig_bound: float = 1e5,
) -> Tuple[ComputerWrapper, ComputerWrapper]:
    """
    Find upper and lower bracket limits for the root finding algorithm

    Args:
        computer (``Callable[[float], float]``): Function that we want to find the root
        loc (``float``, default ``0.0``): location of the root e.g. ``0.95`` for :math:`1-CL_s` value
        low_ini (``float``, default ``1.0``): Initial value for low bracket
        hig_ini (``float``, default ``1.0``): initial value for high bracket
        low_bound (``float``, default ``1e-10``): Stop the execution below this value
        hig_bound (``float``, default ``1e5``): Stop the execution above this value

    Returns:
        ``Tuple[ComputerWrapper, ComputerWrapper]``:
        Returns lower and upper limits for the bracketing within a computer wrapper object.
    """
    assert callable(computer), "Invalid input. Computer must be callable."

    low, hig = low_ini, hig_ini

    low_computer = ComputerWrapper(computer)
    while low_computer(low) > loc:
        low *= 0.5
        if low < low_bound:
            break
    log.debug(f"Low results: {low_computer._results}")

    hig_computer = ComputerWrapper(computer)
    while hig_computer(hig) < loc:
        hig *= 2.0
        if hig > hig_bound:
            break
    log.debug(f"High results: {hig_computer._results}")

    return low_computer, hig_computer


def find_poi_upper_limit(
    maximum_likelihood: Tuple[float, float],
    logpdf: Callable[[float], float],
    maximum_asimov_likelihood: Tuple[float, float],
    asimov_logpdf: Callable[[float], float],
    expected: ExpectationType,
    confidence_level: float = 0.95,
    allow_negative_signal: bool = True,
    low_init: float = 1.0,
    hig_init: float = 1.0,
    expected_pvalue: Literal["nominal", "1sigma", "2sigma"] = "nominal",
    maxiter: int = 10000,
) -> Union[float, List[float]]:
    r"""
    Find upper limit for parameter of interest, :math:`\mu`

    Args:
        maximum_likelihood (``Tuple[float, float]``): Tuple including :math:`\hat\mu`
          and minimum negative log-likelihood.
        logpdf (``Callable[[float], float]``): log-likelihood as function of POI,
          :math:`\log\mathcal{L}(\mu)`
        maximum_asimov_likelihood (``Tuple[float, float]``): Tuple including
          :math:`\hat\mu_A` and minimum negative log-likelihood for Asimov data.
        asimov_logpdf (``Callable[[float], float]``): log-likelihood as function of POI,
          :math:`\log\mathcal{L}_A(\mu)` for Asimov data.
        expected (~spey.ExpectationType): Sets which values the fitting algorithm should
          focus and p-values to be computed.

          * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
            prescriotion which means that the experimental data will be assumed to be the truth
          * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
            post-fit prescriotion which means that the experimental data will be assumed to be
            the truth.
          * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
            prescription which means that the SM will be assumed to be the truth.

        confidence_level (``float``, default ``0.95``): Determines the confidence level of the upper
              limit i.e. the value of :math:`1-CL_s`. It needs to be between ``[0,1]``.
        allow_negative_signal (``bool``, default ``True``): _description_
        low_init (``float``, default ``None``): Lower limit for the search algorithm to start
        hig_init (``float``, default ``None``): Upper limit for the search algorithm to start
        expected_pvalue (``Text``, default ``"nominal"``): In case of :obj:`~spey.ExpectationType.aposteriori`
          and :obj:`~spey.ExpectationType.apriori` expectation, gives the choice to find excluded upper
          limit for statistical deviations as well.

          * ``"nominal"``: only find the upper limit for the central p-value. Returns a single value.
          * ``"1sigma"``: find the upper limit for central p-value and :math:`1\sigma` fluctuation from
            background. Returns 3 values.
          * ``"2sigma"``: find the upper limit for central p-value and :math:`1\sigma` and
            :math:`2\sigma` fluctuation from background. Returns 5 values.

            .. note::

              For ``expected=spey.ExpectationType.observed``, ``expected_pvalue`` argument will
              be overwritten to ``"nominal"``.

        maxiter (``int``, default ``10000``): Maximum iteration limit for the optimiser.

    Returns:
        ``Union[float, List[float]]``:
        In case of nominal values it returns a single value for the upper limit. In case of
        ``expected_pvalue="1sigma"`` or ``expected_pvalue="2sigma"`` it will return a list of
        multiple upper limit values for fluctuations as well as the central value. The
        output order is :math:`-2\sigma` value, :math:`-1\sigma` value, central value,
        :math:`1\sigma` and :math:`2\sigma` value.
    """
    assert expected_pvalue in [
        "nominal",
        "1sigma",
        "2sigma",
    ], f"Unknown pvalue range {expected_pvalue}"
    if expected is ExpectationType.observed:
        expected_pvalue = "nominal"
    test_stat = "q" if allow_negative_signal else "qtilde"

    def computer(poi_test: float, pvalue_idx: int) -> float:
        """Compute 1 - CLs(POI) = `confidence_level`"""
        _, sqrt_qmuA, delta_teststat = compute_teststatistics(
            poi_test,
            maximum_likelihood,
            logpdf,
            maximum_asimov_likelihood,
            asimov_logpdf,
            test_stat,
        )
        pvalue = list(
            map(
                lambda x: 1.0 - x,
                compute_asymptotic_confidence_level(sqrt_qmuA, delta_teststat, test_stat)[
                    0 if expected == ExpectationType.observed else 1
                ],
            )
        )
        return pvalue[pvalue_idx] - confidence_level

    result = []
    index_range = {
        "nominal": [0 if expected is ExpectationType.observed else 2],
        "1sigma": range(1, 4),
        "2sigma": range(0, 5),
    }
    for pvalue_idx in index_range[expected_pvalue]:
        log.debug(f"Running for p-value idx: {pvalue_idx}")
        comp = partial(computer, pvalue_idx=pvalue_idx)
        # Set an upper bound for the computation
        hig_bound = 1e5
        low, hig = find_root_limits(
            comp, loc=0.0, low_ini=low_init, hig_ini=hig_init, hig_bound=hig_bound
        )
        log.debug(f"low: {low[-1]}, hig: {hig[-1]}")

        # Check if its possible to find roots
        if np.sign(low[-1]) * np.sign(hig[-1]) > 0.0:
            log.warning(
                "Can not find the roots of the function, returning `inf`"
                + f"\n hig, low must bracket a root f({low.get_value(-1):.5e})={low[-1]:.5e}, "
                + f"f({hig.get_value(-1):.5e})={hig[-1]:.5e}. "
                + "This is likely due to low number of signal yields."
                * (hig.get_value(-1) >= hig_bound),
            )
            result.append(np.inf)
            continue

        with warnings.catch_warnings(record=True) as w:
            x0, r = scipy.optimize.toms748(
                comp,
                low.get_value(-1),
                hig.get_value(-1),
                k=2,
                xtol=2e-12,
                rtol=1e-4,
                full_output=True,
                maxiter=maxiter,
            )

        log.debug("Warnings:")
        if log.level == logging.DEBUG:
            for warning in w:
                log.debug(f"\t{warning.message}")
            log.debug("<><>" * 10)

        if not r.converged:
            log.warning(f"Optimiser did not converge.\n{r}")
        result.append(x0)
    return result if len(result) > 1 else result[0]



"""Tools for computing third moment expansion"""
import warnings
from typing import Optional, Tuple, Union
import logging
import autograd.numpy as np
from scipy import integrate
from scipy.stats import norm

# pylint: disable=E1101,E1120,W1203
log = logging.getLogger("Spey")


def third_moment_expansion(
    expectation_value: np.ndarray,
    covariance_matrix: np.ndarray,
    third_moment: Optional[np.ndarray] = None,
    return_correlation_matrix: bool = False,
) -> Tuple:
    r"""
    Construct the terms for third moment expansion. For details see :xref:`1809.05548`.

    .. attention::

        This function expects :math:`8\Sigma_{ii}^3 \geq (m^{(3)}_i)^2`. In case its not
        satisfied, `NaN` values will be replaced with zero.

    Args:
        expectation_value (``np.ndarray``): expectation value of the background
        covariance_matrix (``np.ndarray``): covariance matrix
        third_moment (``np.ndarray``): Diagonal components of the third moment, :math:`m^{(3)}`
        return_correlation_matrix (``bool``, default ``False``): If true reconstructs
          and returns correlation matrix.

    Returns:
        ``np.ndarray``:
        A, B, C terms from :xref:`1809.05548` eqns 2.9, 2.10, 2.11. if
        ``return_correlation_matrix`` is ``True`` it also returns correlation matrix.
    """
    cov_diag = np.diag(covariance_matrix)

    if not np.all(8.0 * cov_diag**3 >= third_moment**2):
        log.warning(
            r"Third moments does not satisfy the following condition: $8\Sigma_{ii}^3 \geq (m^{(3)}_i)^2$"
        )
        log.warning("The values that do not satisfy this condition will be set to zero.")

    # arXiv:1809.05548 eq. 2.9
    with warnings.catch_warnings(record=True) as w:
        C = (
            -np.sign(third_moment)
            * np.sqrt(2.0 * cov_diag)
            * np.cos(
                (4.0 * np.pi / 3.0)
                + (1.0 / 3.0)
                * np.arctan(np.sqrt(((8.0 * cov_diag**3) / third_moment**2) - 1.0))
            )
        )
    if len(w) > 0:
        warnings.warn(
            "8 * diag(cov)**3 >= third_moment**2 condition is not satisfied,"
            " setting nan values to zero.",
            category=RuntimeWarning,
        )
        C = np.where(np.isnan(C), 0.0, C)
    log.debug(f"C: {C}")

    # arXiv:1809.05548 eq. 2.10
    B = np.sqrt(cov_diag - 2 * C**2)
    log.debug(f"B: {B}")

    # arXiv:1809.05548 eq. 2.11
    A = expectation_value - C
    log.debug(f"A: {A}")

    # arXiv:1809.05548 eq. 2.12
    eps = 1e-5
    if return_correlation_matrix:
        corr = np.zeros((C.shape[0], C.shape[0]))
        for i in range(corr.shape[0]):
            for j in range(corr.shape[0]):
                ci = C[i] + eps if C[i] >= 0 else C[i] - eps
                cj = C[j] + eps if C[j] >= 0 else C[j] - eps
                cicj = ci * cj
                bibj = B[i] * B[j]

                discr1 = bibj**2
                discr2 = 8 * cicj * covariance_matrix[i, j]
                discr = discr1 + discr2

                corr[i, j] = (np.sqrt(abs(discr)) - bibj) / 4 / cicj

                if i != j:
                    corr[j, i] = corr[i, j]

        log.debug(f"rho: {corr}")
        return A, B, C, corr

    return A, B, C


def compute_third_moments(
    absolute_upper_uncertainties: np.ndarray,
    absolute_lower_uncertainties: np.ndarray,
    return_integration_error: bool = False,
) -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:
    r"""
    Assuming that the uncertainties are modelled as Gaussian, it computes third moments
    using Bifurcated Gaussian with asymmetric uncertainties.

    .. math::

        m^{(3)} = \frac{2}{\sigma^+ + \sigma^-} \left[ \sigma^-\int_{-\infty}^0 x^3
        \mathcal{N}(x|0,\sigma^-)dx + \sigma^+ \int_0^{\infty} x^3
        \mathcal{N}(x|0,\sigma^+)dx \right]

    .. note::

        Recall that expectation value of the :math:`k` th moment of a function :math:`f(x)`
        can be calculated as

        .. math::

            \mathbb{E}[(\mathbf{X} - c)^k] = \int_{-\infty}^\infty(x-c)^kf(x)dx


    .. attention::

        :func:`~spey.backends.default_pdf.third_moment.third_moment_expansion`
        function expects :math:`8\Sigma_{ii}^3 \geq (m^{(3)}_i)^2` since this function is
        constructed with upper and lower uncertainty envelops independent of covariance matrix,
        it does not guarantee that the condition will be satisfied. This depends on the
        covariance matrix.

    Args:
        absolute_upper_uncertainties (``np.ndarray``): absolute value of the upper uncertainties
        absolute_lower_uncertainties (``np.ndarray``): absolute value of the lower uncertainties
        return_integration_error (``bool``, default ``False``): If true returns integration error

    Returns:
        ``Tuple[np.ndarray, np.ndarray]`` or ``np.ndarray``:
        Diagonal elements of the third moments and integration error.
    """

    def compute_x3BifurgatedGaussian(x: float, upper: float, lower: float) -> float:
        Norm = 2.0 / (upper + lower)
        if x >= 0.0:
            # integrate from 0 to inf
            return Norm * upper * x**3 * norm.pdf(x, 0.0, upper)
        # integrate from -inf to 0
        return Norm * lower * x**3 * norm.pdf(x, 0.0, lower)

    third_moment, error = [], []
    for upper, lower in zip(absolute_upper_uncertainties, absolute_lower_uncertainties):
        third_moment_tmp, error_tmp = integrate.quad(
            compute_x3BifurgatedGaussian, -np.inf, np.inf, args=(abs(upper), abs(lower))
        )
        third_moment.append(third_moment_tmp)
        error.append(error_tmp)

    if return_integration_error:
        return np.array(third_moment), np.array(error)
    log.debug(f"Error: {error}")

    return np.array(third_moment)



"""Statistical Model wrapper class"""
import logging
from typing import Any, Callable, Dict, List, Optional, Text, Tuple, Union

import numpy as np

from spey.base.backend_base import BackendBase
from spey.base.hypotest_base import HypothesisTestingBase
from spey.optimizer.core import fit
from spey.system.exceptions import (
    CombinerNotAvailable,
    MethodNotAvailable,
    UnknownCrossSection,
)
from spey.utils import ExpectationType

__all__ = ["StatisticalModel", "statistical_model_wrapper"]


def __dir__():
    return __all__


log = logging.getLogger("Spey")

# pylint: disable=W1203


class StatisticalModel(HypothesisTestingBase):
    r"""
    Statistical model base. This class wraps around the various statistical model backends available
    through `spey`'s plugin system. Each backend has to inherit :class:`~spey.BackendBase` which sets
    certain requirements on the available functionality to be used for hypothesis testing. These
    requirements are such as accessibility to log-likelihood, :math:`\log\mathcal{L}`, it's derivative
    with respect to :math:`\mu` and nuisance parameters, :math:`\partial_\theta\log\mathcal{L}`,
    its Hessian and Assimov data generation. Depending on availablility :class:`~spey.StatisticalModel`
    will take propriate action to perform requested computation. The goal of this class is to collect
    all different backends under same roof in order to perform combination of different likelihood
    recipies.

    Args:
        backend (~spey.BackendBase): Statistical model backend
        analysis (``Text``): Unique identifier of the statistical model. This attribue will be used
          for book keeping purposes.
        xsection (``float``, default ``np.nan``): cross section, unit is determined by the user.
          Cross section value is only used for computing upper limit on excluded cross-section value.
        ntoys (``int``, default ``1000``): Number of toy samples for hypothesis testing. (Only used for
          toy-based hypothesis testing)

    Raises:
        :obj:`AssertionError`: If the given backend does not inherit :class:`~spey.BackendBase`

    Returns:
        ~spey.StatisticalModel:
        General statistical model object that wraps around different likelihood prescriptions.
    """

    __slots__ = ["_backend", "xsection", "analysis"]

    def __init__(
        self,
        backend: BackendBase,
        analysis: Text,
        xsection: float = np.nan,
        ntoys: int = 1000,
    ):
        assert isinstance(backend, BackendBase), "Invalid backend"
        self._backend: BackendBase = backend
        self.xsection: float = xsection
        """Value of the cross section, unit is defined by the user."""
        self.analysis: Text = analysis
        """Unique identifier as analysis name"""
        super().__init__(ntoys=ntoys)

    def __repr__(self):
        calc = f"calculators={self.available_calculators}"

        if np.isnan(self.xsection):
            return (
                f"StatisticalModel(analysis='{self.analysis}', backend={self.backend_type}, "
                + f"{calc})"
            )

        return (
            f"StatisticalModel(analysis='{self.analysis}', "
            f"xsection={self.xsection:.3e} [au], backend={self.backend_type}, {calc})"
        )

    @property
    def backend(self) -> BackendBase:
        """Accessor to the backend"""
        return self._backend

    @property
    def backend_type(self) -> Text:
        """Return type of the backend"""
        return getattr(self.backend, "name", self.backend.__class__.__name__)

    @property
    def is_asymptotic_calculator_available(self) -> bool:
        """Check if Asymptotic calculator is available for the backend"""
        return self.backend.expected_data != BackendBase.expected_data or (
            self.backend.asimov_negative_loglikelihood
            != BackendBase.asimov_negative_loglikelihood
            and self.backend.minimize_asimov_negative_loglikelihood
            != BackendBase.minimize_asimov_negative_loglikelihood
        )

    @property
    def is_toy_calculator_available(self) -> bool:
        """Check if Toy calculator is available for the backend"""
        return self.backend.get_sampler != BackendBase.get_sampler

    @property
    def is_chi_square_calculator_available(self) -> bool:
        """Check if chi-square calculator is available for the backend"""
        return True

    @property
    def available_calculators(self) -> List[Text]:
        """
        Retruns available calculator names i.e. ``'toy'``,
        ``'asymptotic'`` and ``'chi_square'``.
        """
        calc = ["toy"] * self.is_toy_calculator_available
        calc += ["asymptotic"] * self.is_asymptotic_calculator_available
        calc += ["chi_square"] * self.is_chi_square_calculator_available
        return calc

    @property
    def is_alive(self) -> bool:
        """Returns True if at least one bin has non-zero signal yield."""
        return self.backend.is_alive

    def excluded_cross_section(
        self, expected: ExpectationType = ExpectationType.observed
    ) -> float:
        """
        Compute excluded cross section value at 95% CL

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

        Raises:
            ~spey.system.exceptions.UnknownCrossSection: If the cross-section is ``nan``.

        Returns:
            ``float``:
            Returns the upper limit at 95% CL on cross section value where the unit is defined
            by the user.
        """
        if np.isnan(self.xsection):
            raise UnknownCrossSection("Cross-section value has not been initialised.")

        return (
            self.poi_upper_limit(expected=expected, confidence_level=0.95) * self.xsection
        )

    @property
    def s95exp(self) -> float:
        """
        Compute excluded cross section value at 95% CL with :obj:`~spey.ExpectationType.apriori`
        expectation. See :func:`~spey.StatisticalModel.excluded_cross_section`
        for reference.

        Raises:
            ~spey.system.exceptions.UnknownCrossSection: If the cross-section is ``nan``.
        """
        return self.excluded_cross_section(ExpectationType.apriori)

    @property
    def s95obs(self) -> float:
        """
        Compute excluded cross section value at 95% CL with :obj:`~spey.ExpectationType.observed`
        expectation. See :func:`~spey.StatisticalModel.excluded_cross_section`
        for reference.

        Raises:
            ~spey.system.exceptions.UnknownCrossSection: If the cross-section is ``nan``.
        """
        return self.excluded_cross_section(ExpectationType.observed)

    def prepare_for_fit(
        self,
        data: Optional[Union[List[float], np.ndarray]] = None,
        expected: ExpectationType = ExpectationType.observed,
        allow_negative_signal: Optional[bool] = True,
        **kwargs,
    ) -> Dict:
        r"""
        Prepare backend for the optimiser.

        Args:
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.

        Returns:
            ``Dict``:
            Dictionary of necessary toolset for the fit. objective function, ``"func"``, use gradient
            boolean, ``"do_grad"`` and function to compute negative log-likelihood with given
            fit parameters, ``"nll"``.
        """
        do_grad = kwargs.pop("do_grad", True)
        try:
            objective_and_grad = self.backend.get_objective_function(
                expected=expected, data=data, do_grad=do_grad
            )
        except NotImplementedError:
            log.debug("Gradient is not available, will not be included in computation.")
            do_grad = False
            objective_and_grad = self.backend.get_objective_function(
                expected=expected, data=data, do_grad=do_grad
            )

        constraints = kwargs.pop("constraints", [])
        if hasattr(self.backend, "constraints"):
            for constraint in self.backend.constraints:
                constraints.append(constraint)

        return {
            "func": objective_and_grad,
            "do_grad": do_grad,
            "model_configuration": self.backend.config(
                allow_negative_signal=allow_negative_signal
            ),
            "logpdf": self.backend.get_logpdf_func(expected=expected, data=data),
            "constraints": constraints,
            **kwargs,
        }

    def likelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        return_nll: bool = True,
        data: Optional[Union[List[float], np.ndarray]] = None,
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> float:
        r"""
        Compute the likelihood of the statistical model at a fixed parameter of interest

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest or signal strength,
              :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            return_nll (``bool``, default ``True``): If ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit. If
              ``None`` data will be set according to ``expected`` input.
            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.

        Returns:
            ``float``:
            Likelihood of the statistical model at a fixed signal strength.
        """
        fit_opts = self.prepare_for_fit(expected=expected, data=data, **kwargs)

        if (
            fit_opts["model_configuration"].npar == 1
            and fit_opts["model_configuration"].poi_index is not None
        ):
            logpdf = fit_opts["logpdf"]([poi_test])
        else:
            logpdf, _ = fit(
                **fit_opts,
                initial_parameters=init_pars,
                bounds=par_bounds,
                fixed_poi_value=poi_test,
            )

        return -logpdf if return_nll else np.exp(logpdf)

    def generate_asimov_data(
        self,
        expected: ExpectationType = ExpectationType.observed,
        test_statistic: Text = "qtilde",
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> List[float]:
        r"""
        Generate Asimov data for the statistical model. This function generates a set of parameters
        (nuisance and poi i.e. :math:`\theta` and :math:`\mu`) with respect to ``test_statistic`` input
        which determines the value of :math:`\mu` i.e. if ``test_statistic="q0"`` :math:`\mu=1` and 0 for
        anything else. The objective function is used to optimize the statistical model to find the fit
        parameters for fixed poi optimisation. Then fit parameters are used to retreive the expected
        data through :func:`~spey.BackendBase.expected_data` function.

        Args:
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistic (``Text``, default ``"qtilde"``): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence ``allow_negative_signal``
                    assumed to be ``False``. If this function has been executed by user, ``spey``
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through ``spey`` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.

        Returns:
            ``List[float]``:
            Asimov data
        """
        fit_opts = self.prepare_for_fit(
            expected=expected,
            allow_negative_signal=test_statistic in ["q", "qmu"],
            **kwargs,
        )

        _, fit_pars = fit(
            **fit_opts,
            initial_parameters=init_pars,
            bounds=par_bounds,
            fixed_poi_value=1.0 if test_statistic == "q0" else 0.0,
        )
        log.debug(f"fit parameters: {fit_pars}")

        return self.backend.expected_data(fit_pars)

    def asimov_likelihood(
        self,
        poi_test: float = 1.0,
        expected: ExpectationType = ExpectationType.observed,
        return_nll: bool = True,
        test_statistics: Text = "qtilde",
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> float:
        r"""
        Compute likelihood of the statistical model generated with the Asimov data.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest, :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            return_nll (``bool``, default ``True``): If ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            test_statistics (``Text``, default ``"qtilde"``): test statistics.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence ``allow_negative_signal``
                    assumed to be ``False``. If this function has been executed by user, ``spey``
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through ``spey`` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

              The choice of ``test_statistics`` will effect the generation of the Asimov data where
              the fit is performed via :math:`\mu=1` if ``test_statistics="q0"`` and :math:`\mu=0`
              for others. Note that this :math:`\mu` does not correspond to the ``poi_test`` input
              of this function but it determines how Asimov data is generated.
            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.

        Returns:
            ``float``:
            likelihood computed for asimov data
        """
        return self.likelihood(
            poi_test=poi_test,
            expected=expected,
            return_nll=return_nll,
            data=self.generate_asimov_data(
                expected=expected,
                test_statistic=test_statistics,
                init_pars=init_pars,
                par_bounds=par_bounds,
                **kwargs,
            ),
            init_pars=init_pars,
            par_bounds=par_bounds,
            **kwargs,
        )

    def maximize_likelihood(
        self,
        return_nll: Optional[bool] = True,
        expected: Optional[ExpectationType] = ExpectationType.observed,
        allow_negative_signal: Optional[bool] = True,
        data: Optional[Union[List[float], np.ndarray]] = None,
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> Tuple[float, float]:
        r"""
        Find the maximum of the likelihood.

        Args:
            return_nll (``bool``, default ``True``): If  ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            allow_negative_signal (``bool``, default ``True``): If ``True`` :math:`\hat\mu`
              value will be allowed to be negative.
            data (``Union[List[float], np.ndarray]``, default ``None``): input data that to fit. If
              ``None`` data will be set according to ``expected`` input.
            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default  ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.

        Returns:
            ``Tuple[float, float]``:
            :math:`\hat\mu` value and maximum value of the likelihood.
        """
        fit_opts = self.prepare_for_fit(
            expected=expected,
            allow_negative_signal=allow_negative_signal,
            data=data,
            **kwargs,
        )

        logpdf, fit_param = fit(
            **fit_opts, initial_parameters=init_pars, bounds=par_bounds
        )
        log.debug(f"fit parameters: {fit_param}")

        muhat = fit_param[self.backend.config().poi_index]
        return muhat, -logpdf if return_nll else np.exp(logpdf)

    def maximize_asimov_likelihood(
        self,
        return_nll: bool = True,
        expected: ExpectationType = ExpectationType.observed,
        test_statistics: Text = "qtilde",
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> Tuple[float, float]:
        r"""
        Find the maximum of the likelihood which computed with respect to Asimov data.

        Args:
            return_nll (``bool``, default ``True``): If ``True``, returns negative log-likelihood value.
              if ``False`` returns likelihood value.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.

            test_statistics (``Text``, default ``"qtilde"``): test statistic.

              * ``'qtilde'``: (default) performs the calculation using the alternative test statistic,
                :math:`\tilde{q}_{\mu}`, see eq. (62) of :xref:`1007.1727`
                (:func:`~spey.hypothesis_testing.test_statistics.qmu_tilde`).

                .. warning::

                    Note that this assumes that :math:`\hat\mu\geq0`, hence ``allow_negative_signal``
                    assumed to be ``False``. If this function has been executed by user, ``spey``
                    assumes that this is taken care of throughout the external code consistently.
                    Whilst computing p-values or upper limit on :math:`\mu` through ``spey`` this
                    is taken care of automatically in the backend.

              * ``'q'``: performs the calculation using the test statistic :math:`q_{\mu}`, see
                eq. (54) of :xref:`1007.1727` (:func:`~spey.hypothesis_testing.test_statistics.qmu`).
              * ``'q0'``: performs the calculation using the discovery test statistic, see eq. (47)
                of :xref:`1007.1727` :math:`q_{0}` (:func:`~spey.hypothesis_testing.test_statistics.q0`).

            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.

        Returns:
            ``Tuple[float, float]``:
            :math:`\hat\mu` value and maximum value of the likelihood.
        """
        allow_negative_signal: bool = True if test_statistics in ["q", "qmu"] else False

        return self.maximize_likelihood(
            return_nll=return_nll,
            expected=expected,
            allow_negative_signal=allow_negative_signal,
            data=self.generate_asimov_data(
                expected=expected,
                test_statistic=test_statistics,
                init_pars=init_pars,
                par_bounds=par_bounds,
                **kwargs,
            ),
            init_pars=init_pars,
            par_bounds=par_bounds,
            **kwargs,
        )

    def fixed_poi_sampler(
        self,
        poi_test: float,
        size: Optional[int] = None,
        expected: ExpectationType = ExpectationType.observed,
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> Union[np.ndarray, Callable[[int], np.ndarray]]:
        r"""
        Sample data from the statistical model with fixed parameter of interest.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest or signal strength,
              :math:`\mu`.
            size (``int``, default ``None``): sample size. If ``None`` a callable function
              will be returned which takes sample size as input.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.

        Raises:
            ~spey.system.exceptions.MethodNotAvailable: If bacend does not have sampler implementation.

        Returns:
            ``Union[np.ndarray, Callable[[int], np.ndarray]]``:
            Sampled data with shape of ``(size, number of bins)`` or callable function to sample from
            directly.
        """
        fit_opts = self.prepare_for_fit(expected=expected, **kwargs)

        if (
            fit_opts["model_configuration"].npar == 1
            and fit_opts["model_configuration"].poi_index is not None
        ):
            fit_param = np.array(
                poi_test if isinstance(poi_test, (list, np.ndarray)) else [poi_test]
            )
        else:
            _, fit_param = fit(
                **fit_opts,
                initial_parameters=init_pars,
                bounds=par_bounds,
                fixed_poi_value=poi_test,
            )

        log.debug(f"fit parameters: {fit_param}")
        try:
            sampler = self.backend.get_sampler(fit_param)
        except NotImplementedError as exc:
            raise MethodNotAvailable(
                f"{self.backend_type} backend does not have sampling capabilities"
            ) from exc

        return sampler(size) if isinstance(size, int) else sampler

    def sigma_mu_from_hessian(
        self,
        poi_test: float,
        expected: ExpectationType = ExpectationType.observed,
        init_pars: Optional[List[float]] = None,
        par_bounds: Optional[List[Tuple[float, float]]] = None,
        **kwargs,
    ) -> float:
        r"""
        Compute variance of :math:`\mu` from inverse Hessian. See eq. (27-28) in :xref:`1007.1727`.

        Args:
            poi_test (``float``, default ``1.0``): parameter of interest or signal strength,
              :math:`\mu`.
            expected (~spey.ExpectationType): Sets which values the fitting algorithm should focus and
              p-values to be computed.

              * :obj:`~spey.ExpectationType.observed`: Computes the p-values with via post-fit
                prescriotion which means that the experimental data will be assumed to be the truth
                (default).
              * :obj:`~spey.ExpectationType.aposteriori`: Computes the expected p-values with via
                post-fit prescriotion which means that the experimental data will be assumed to be
                the truth.
              * :obj:`~spey.ExpectationType.apriori`: Computes the expected p-values with via pre-fit
                prescription which means that the SM will be assumed to be the truth.
            init_pars (``List[float]``, default ``None``): initial parameters for the optimiser
            par_bounds (``List[Tuple[float, float]]``, default ``None``): parameter bounds for
              the optimiser.
            kwargs: keyword arguments for the optimiser.
        Raises:
            ~spey.system.exceptions.MethodNotAvailable: If bacend does not have Hessian implementation.

        Returns:
            ``float``:
            variance on parameter of interest.
        """
        try:
            hessian_func = self.backend.get_hessian_logpdf_func(expected=expected)
        except NotImplementedError as exc:
            raise MethodNotAvailable(
                f"{self.backend_type} backend does not have Hessian definition."
            ) from exc

        fit_opts = self.prepare_for_fit(expected=expected, **kwargs)
        _ = fit_opts.pop("logpdf")

        _, fit_param = fit(
            **fit_opts,
            initial_parameters=init_pars,
            bounds=par_bounds,
            fixed_poi_value=poi_test,
        )
        log.debug(f"fit parameters: {fit_param}")

        hessian = -1.0 * hessian_func(fit_param)
        log.debug(f"full hessian: {hessian}")

        poi_index = self.backend.config().poi_index
        return np.sqrt(np.linalg.inv(hessian)[poi_index, poi_index])

    def combine(self, other, **kwargs):
        """
        Combination routine between two statistical models.

        .. note::

            This function's availability is backend dependent.

        Args:
            other (:obj:`~spey.StatisticalModel`): Statistical model to be combined with
              this model
            kwargs: backend specific arguments.

        Raises:
            :obj:`~spey.system.exceptions.CombinerNotAvailable`: If this statistical model
              does not have a combination routine implementation.
            ``AssertionError``: If the combination routine in the backend does not return
              a :obj:`~spey.BackendBase` object.

        Returns:
            :obj:`~spey.StatisticalModel`:
            Returns a new combined statistical model.
        """
        try:
            combined = self.backend.combine(other.backend, **kwargs)
            assert isinstance(combined, BackendBase), "Invalid combination operation."

            return StatisticalModel(
                backend=combined, analysis=f"combine[{self.analysis}, {other.analysis}]"
            )

        except NotImplementedError as err:
            raise CombinerNotAvailable(
                f"{self.backend_type} backend does not have a combination routine."
            ) from err

    def __matmul__(self, other):
        """
        Combination routine between two statistical models.
        See :func:`~spey.StatisticalModel.combine` function for details.
        """
        return self.combine(other)


def statistical_model_wrapper(
    func: BackendBase,
) -> Callable[[Any], StatisticalModel]:
    """
    Backend wrapper for :class:`~spey.StatisticalModel`. This function allows a universal
    integration of each backend to the :obj:`spey` environment. :func:`~spey.get_backend` function
    automatically wraps the backend with :func:`~spey.statistical_model_wrapper` before returning
    the object.

    Args:
        func (~spey.BackendBase): Desired backend to be used for statistical analysis.

    Raises:
        :obj:`AssertionError`: If the input function does not inherit :obj:`~spey.BackendBase`

    Returns:
        ``Callable[[Any], StatisticalModel]``:
        Wrapper that takes the following inputs

        * **args**: Backend specific arguments.
        * **analysis** (``Text``, default ``"__unknown_analysis__"``): Unique identifier of the
          statistical model. This attribue will be used for book keeping purposes.
        * **xsection** (``float``, default ``nan``): cross section, unit is determined by the
          user. Cross section value is only used for computing upper limit on excluded
          cross-section value.
        * **ntoys** (``int``, default ``1000``): Number of toy samples for hypothesis testing.
            (Only used for toy-based hypothesis testing)
        * **other keyword arguments**: Backend specific keyword inputs.
    """

    def wrapper(
        *args,
        analysis: Text = "__unknown_analysis__",
        xsection: float = np.nan,
        ntoys: int = 1000,
        **kwargs,
    ) -> StatisticalModel:
        """
        Statistical Model Backend wrapper.

        Args:
            args: Backend specific arguments.
            analysis (``Text``, default ``"__unknown_analysis__"``): Unique identifier of the
              statistical model. This attribue will be used for book keeping purposes.
            xsection (``float``, default ``nan``): cross section, unit is determined by the
              user. Cross section value is only used for computing upper limit on excluded
              cross-section value.
            ntoys (``int``, default ``1000``): Number of toy samples for hypothesis testing.
              (Only used for toy-based hypothesis testing)
            kwargs: Backend specific keyword inputs.

        Raises:
            :obj:`AssertionError`: If the model input does not inherit :class:`~spey.DataBase`.

        Returns:
            ~spey.StatisticalModel:
            Backend wraped with statistical model interface.
        """
        return StatisticalModel(
            backend=func(*args, **kwargs),
            analysis=analysis,
            xsection=xsection,
            ntoys=ntoys,
        )

    docstring = getattr(func, "__doc__", "no docstring available")
    if docstring is None:
        docstring = "Documentation is not available..."

    wrapper.__doc__ += (
        "\n\t"
        + "<>" * 30
        + "\n\n\t Current statistical model backend properties:\n"
        + docstring.replace("\n", "\n\t")
        + "\n"
    )

    return wrapper
